# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


import typing as T
from dataclasses import dataclass, field

from uidom.dom.htmlelement import HTMLElement
from valio import StringValidator, Validator

__all__ = [
    # integer html
    "IntegerLabel",
    "IntegerLegend",
    "IntegerInput",
    "IntegerNumberInput",
    "IntegerRangeInput",
    "TelephoneInput",
    "IntegerField",
]

@Validator.register
@dataclass(eq=False)
class IntegerLabel(HTMLElement):
    label: str = StringValidator(logger=False, debug=True)

    def __post_init__(self, *args, **kwargs):
        super(IntegerLabel, self).__init__(*args, label=self.label, **kwargs)

    def __render__(self, label, *args, **kwargs):
        return self.html_tags.label(label, *args, **kwargs)



class IntegerLabelValidator(Validator):
    annotation = T.Union[IntegerLabel, None]


@dataclass(eq=False)
class IntegerLegend(HTMLElement):
    label: str = StringValidator(logger=False, debug=True)

    def __post_init__(self, *args, **kwargs):
        super(IntegerLegend, self).__init__(label=self.label, *args, *kwargs)

    def __render__(self, label, *args, **kwargs):
        return self.html_tags.legend(label, *args, **kwargs)

min_field = StringValidator(logger=False, debug=True, name="min")
max_field = StringValidator(logger=False, debug=True, name="max")

@dataclass(eq=False)
class IntegerInput(HTMLElement):

    @max_field.add_validator
    @min_field.add_validator
    def cast2int(self, value):
        if value not in ["null", None]:
            try:
                int(value)
            except TypeError as e:
                raise ValueError(f"{value} is not an integer value") from e

    name: str = StringValidator(logger=False, debug=True)
    placeholder: str = StringValidator(logger=False, debug=True)
    type: str = StringValidator(in_choice=["number", "range", "tel"],
                                logger=False, debug=True, default="number")
    min: str = min_field
    max: str = max_field
    pattern: str = StringValidator(logger=False, debug=True, default=None)

    def __post_init__(self, *args, **kwargs):
        super(IntegerInput, self).__init__(*args,
                                           name=self.name,
                                           placeholder=self.placeholder,
                                           type=self.type,
                                           min=self.min,
                                           max=self.max,
                                           pattern=self.pattern,
                                           **kwargs
                                           )

    def __render__(
            self,
            *args,
            name,
            placeholder="",
            type,
            min,
            max,
            pattern,
            **kwargs
    ):

        return self.html_tags.input_(
            *args,
            name=name,
            placeholder=placeholder,
            type=type,
            min=min if min not in ["null", None] else False,
            max=max if max not in ["null", None] else False,
            pattern=pattern if pattern not in ["null", None] else False,
            **kwargs,
        )


Validator.register(IntegerInput)  # noqa


class IntegerInputValidator(Validator):
    annotation = T.Union[IntegerInput, None]


@dataclass(eq=False)
class IntegerNumberInput(HTMLElement):

    name: str
    placeholder: str
    min: T.Union[str, None] = field(default=None)
    max: T.Union[str, None] = field(default=None)
    pattern: T.Union[str, None] = field(default=None)

    def __post_init__(self, *args, **kwargs):
        super(IntegerNumberInput, self).__init__(*args,
                                                 name=self.name,
                                                 placeholder=self.placeholder,
                                                 min=self.min,
                                                 max=self.max,
                                                 pattern=self.pattern,
                                                 **kwargs
                                                 )

    def __render__(
            self,
            *args,
            name,
            placeholder="",
            min,
            max,
            pattern="null",
            **kwargs
    ):
        return IntegerInput(
            name=name,
            placeholder=placeholder,
            type="number",
            min=min,
            max=max,
            pattern=pattern,
        )


Validator.register(IntegerNumberInput)  # noqa


class IntegerNumberInputValidator(Validator):
    annotation = T.Union[IntegerNumberInput, None]


@dataclass(eq=False)
class IntegerRangeInput(HTMLElement):

    name: str
    placeholder: str
    min: T.Union[str, None] = field(default=None)
    max: T.Union[str, None] = field(default=None)
    pattern: T.Union[str, None] = field(default=None)

    def __post_init__(self, *args, **kwargs):
        super(IntegerRangeInput, self).__init__(*args,
                                                name=self.name,
                                                placeholder=self.placeholder,
                                                min=self.min,
                                                max=self.max,
                                                pattern=self.pattern,
                                                **kwargs
                                                )

    def __render__(
            self,
            *args,
            name,
            placeholder="",
            min="null",
            max="null",
            pattern="null",
            **kwargs
    ):
        return IntegerInput(
            type="range",
            name=name,
            placeholder=placeholder,
            min=min,
            max=max,
            pattern=pattern
        )


Validator.register(IntegerRangeInput)  # noqa


class IntegerRangeInputValidator(Validator):
    annotation = T.Union[IntegerRangeInput, None]


@dataclass(eq=False)
class TelephoneInput(HTMLElement):

    name: str
    placeholder: str
    min: T.Union[str, None] = field(default=None)
    max: T.Union[str, None] = field(default=None)
    pattern: T.Union[str, None] = field(default=None)

    def __post_init__(self, *args, **kwargs):
        super(TelephoneInput, self).__init__(*args,
                                             name=self.name,
                                             placeholder=self.placeholder,
                                             min=self.min,
                                             max=self.max,
                                             pattern=self.pattern,
                                             **kwargs
                                             )

    def __render__(
            self,
            *args,
            name,
            placeholder="",
            min,
            max,
            pattern,
            **kwargs
    ):
        return IntegerInput(
            type="tel",
            name=name,
            placeholder=placeholder,
            min=min,
            max=max,
            pattern=pattern,
        )


@dataclass(eq=False)
class IntegerField(HTMLElement):
    labeled = IntegerLabelValidator(logger=False, debug=True)
    input = IntegerInputValidator(logger=False, debug=True)

    label: str
    name: str
    placeholder: str
    type: str

    def __post_init__(self, *args, **kwargs):
        super(IntegerField, self).__init__(*args,
                                           label=self.label,
                                           name=self.name,
                                           placeholder=self.placeholder,
                                           type=self.type,
                                           **kwargs
                                           )

    def __render__(self, *args, label, name, placeholder, type, **kwargs):
        self.labeled = IntegerLabel(label=label)
        self.input = IntegerInput(name=name, placeholder=placeholder, type=type)
        return self.html_tags.div(*args, self.labeled, self.input, **kwargs)


@dataclass(eq=False)
class IntegerNumberField(HTMLElement):
    labeled = IntegerLabelValidator(logger=False, debug=True)
    input = IntegerNumberInputValidator(logger=False, debug=True)

    label: str
    name: str
    placeholder: str

    def __post_init__(self, *args, **kwargs):
        super(IntegerNumberField, self).__init__(*args,
                                                 label=self.label,
                                                 name=self.name,
                                                 placeholder=self.placeholder,
                                                 **kwargs
                                                 )

    def __render__(self, *args, label, name, placeholder, **kwargs):
        self.labeled = IntegerLabel(label=label)
        self.input = IntegerNumberInput(name=name, placeholder=placeholder)
        return self.html_tags.div(*args, self.labeled, self.input, **kwargs)


@dataclass(eq=False)
class IntegerRangeField(HTMLElement):
    labeled = IntegerLabelValidator(logger=False, debug=True)
    input = IntegerRangeInputValidator(logger=False, debug=True)

    label: str
    name: str
    placeholder: str

    def __post_init__(self, *args, **kwargs):
        super(IntegerRangeField, self).__init__(*args,
                                                label=self.label,
                                                name=self.name,
                                                placeholder=self.placeholder,
                                                **kwargs
                                                )

    def __render__(self, *args, label, name, placeholder, **kwargs):
        self.labeled = IntegerLabel(label=label)
        self.input = IntegerRangeInput(name=name, placeholder=placeholder)
        return self.html_tags.div(*args, self.labeled, self.input, **kwargs)


if __name__ == '__main__':
    from uidom.dom import SubmitButton, form
    quantity = IntegerNumberField(label="Quantity", name="quantity", placeholder="quantity")
    quantity.input["id"] = quantity.labeled["for"] = "quantity-id"
    price = IntegerNumberField(label="Price", name="price", placeholder="price")
    price.input["id"] = price.labeled["for"] = "price-id"
    button = SubmitButton("Submit", value="submit")
    forms = form(IntegerLegend("Product"), price, quantity, button)
    print(forms)
    print(IntegerNumberInput(name="price", placeholder="Price", min="2"))



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


import typing as T
from dataclasses import dataclass, field

from uidom.dom.htmlelement import HTMLElement
from valio import StringValidator, Validator

__all__ = [
    # char html
    "CharLabel",
    "CharLegend",
    "CharInput",
    "TextInput",
    "PasswordInput",
    "HiddenInput",
    "SearchInput",
    "CharField",
    "CharFieldSet"
]


@dataclass(eq=False)
class CharLabel(HTMLElement):
    label: str = StringValidator(logger=False, debug=True)

    def __post_init__(self, *args, **kwargs):
        super(CharLabel, self).__init__(label=self.label, *args, **kwargs)

    def __render__(self, label: str, *args, **kwargs):  # noqa
        return self.html_tags.label(label, *args, **kwargs)


class CharLabelValidator(Validator):
    annotation = T.Union[CharLabel, None]


@dataclass(eq=False)
class CharLegend(HTMLElement):
    label: str = StringValidator(logger=False, debug=True)

    def __post_init__(self, *args, **kwargs):
        super(CharLegend, self).__init__(label=self.label, *args, **kwargs)

    def __render__(self, label, *args, **kwargs):
        return self.html_tags.legend(label, *args, **kwargs)


min_length_field = StringValidator(logger=False, debug=True, name="min_length")
max_length_field = StringValidator(logger=False, debug=True, name="max_length")
spell_check_field = StringValidator(in_choice=["true", "false"], 
                                    logger=False, debug=True, name="spell_check", 
                                    default="true")


@dataclass(eq=False)
class CharInput(HTMLElement):

    @max_length_field.add_validator
    @min_length_field.add_validator
    def cast2int(self, value):
        if value not in ["null", None]:
            try:
                int(value)
            except TypeError as e:
                raise ValueError(f"{value} is not an int value") from e

    @spell_check_field.add_post_validator
    def password_spell_check_false(self, value):
        if self.type in ["password", "hidden"]:
            return "false"
        return value

    name: T.Union[str, None] = StringValidator(logger=False, debug=True)
    type: T.Union[str, None] = StringValidator(
        in_choice=["text", "password", "hidden", "email", "search"],
        debug=True, logger=False, default="text"
        )

    placeholder: T.Union[str, None] = StringValidator(logger=False, debug=True)
    spell_check: T.Union[str, None] = spell_check_field
    min_length: T.Union[str, None] = min_length_field
    max_length: T.Union[str, None] = max_length_field
    pattern: T.Union[str, None] = StringValidator(logger=False, debug=True, default="null")

    def __post_init__(self, *args, **kwargs):
        super(CharInput, self).__init__(*args,
                                        name=self.name,
                                        placeholder=self.placeholder,
                                        spell_check=self.spell_check,
                                        type=self.type,
                                        min_length=self.min_length,
                                        max_length=self.max_length,
                                        pattern=self.pattern,
                                        **kwargs)

    def __render__(
            self,
            *args,
            name: str = None,
            placeholder: str = "",
            spell_check: str,
            type: str,
            min_length: str,
            max_length: str,
            pattern: str,
            **kwargs
    ):  
        
        return self.html_tags.input_(
            *args,
            type=type,
            name=name if name is not None else False,
            placeholder=placeholder if placeholder not in ["", None] else False,
            minlength=min_length if min_length not in ["null", None] else False,
            maxlength=max_length if max_length not in ["null", None] else False,
            spellcheck=spell_check if spell_check not in ["null", None] else False,
            pattern=pattern if pattern not in ["null", None] else False,
            **kwargs,
        )


class CharInputValidator(Validator):
    annotation = T.Union[CharInput, None]


@dataclass(eq=False)
class TextInput(HTMLElement):

    name: str
    placeholder: str
    spell_check: T.Union[str, None] = field(default=None)
    min_length: T.Union[str, None] = field(default=None)
    max_length: T.Union[str, None] = field(default=None)
    pattern: T.Union[str, None] = field(default=None)

    def __post_init__(self, *args, **kwargs):
        super(TextInput, self).__init__(*args,
                                        name=self.name,
                                        placeholder=self.placeholder,
                                        spell_check=self.spell_check,
                                        min_length=self.min_length,
                                        max_length=self.max_length,
                                        pattern=self.pattern,
                                        **kwargs)

    def __render__(
            self,
            *args,
            name: str = None,
            placeholder: str = "",
            spell_check: str = "true",
            min_length: str,
            max_length: str = "null",
            pattern: str = "null",
            **kwargs
    ):
        return CharInput(
            name=name,
            placeholder=placeholder,
            type="text",
            spell_check=spell_check,
            min_length=min_length,
            max_length=max_length,
            pattern=pattern,
        )


@dataclass(eq=False)
class PasswordInput(HTMLElement):
    name: str
    placeholder: str
    spell_check: T.Union[str, None] = field(default=None)
    min_length: T.Union[str, None] = field(default=None)
    max_length: T.Union[str, None] = field(default=None)
    pattern: T.Union[str, None] = field(default=None)

    def __post_init__(self, *args, **kwargs):
        super(PasswordInput, self).__init__(*args,
                                            name=self.name,
                                            placeholder=self.placeholder,
                                            spell_check=self.spell_check,
                                            min_length=self.min_length,
                                            max_length=self.max_length,
                                            pattern=self.pattern,
                                            **kwargs)

    def __render__(
            self,
            *args,
            name,
            placeholder,
            spell_check,
            min_length,
            max_length,
            pattern,
            **kwargs
    ):
        return CharInput(
            *args,
            name=name,
            placeholder=placeholder,
            type="password",
            spell_check=spell_check,
            min_length=min_length,
            max_length=max_length,
            pattern=pattern,
            **kwargs
        )


@dataclass(eq=False)
class HiddenInput(HTMLElement):
    name: str
    placeholder: str
    spell_check: T.Union[str, None] = field(default=None)
    min_length: T.Union[str, None] = field(default=None)
    max_length: T.Union[str, None] = field(default=None)
    pattern: T.Union[str, None] = field(default=None)

    def __post_init__(self, *args, **kwargs):
        super(HiddenInput, self).__init__(*args,
                                          name=self.name,
                                          placeholder=self.placeholder,
                                          spell_check=self.spell_check,
                                          min_length=self.min_length,
                                          max_length=self.max_length,
                                          pattern=self.pattern,
                                          **kwargs)

    def __render__(
            self,
            *args,
            name: str = None,
            placeholder: str = "",
            spell_check: str = "true",
            min_length: str = "null",
            max_length: str = "null",
            pattern: str = "null",
            **kwargs
    ):
        return CharInput(
            *args,
            name=name,
            placeholder=placeholder,
            type="hidden",
            spell_check=spell_check,
            min_length=min_length,
            max_length=max_length,
            pattern=pattern,
            **kwargs
        )


@dataclass(eq=False)
class EmailInput(HTMLElement):
    name: str
    placeholder: str
    spell_check: T.Union[str, None] = field(default=None)
    min_length: T.Union[str, None] = field(default=None)
    max_length: T.Union[str, None] = field(default=None)
    pattern: T.Union[str, None] = field(default=None)

    def __post_init__(self, *args, **kwargs):
        super(EmailInput, self).__init__(*args,
                                         name=self.name,
                                         placeholder=self.placeholder,
                                         spell_check=self.spell_check,
                                         min_length=self.min_length,
                                         max_length=self.max_length,
                                         pattern=self.pattern,
                                         **kwargs)

    def __render__(
            self,
            *args,
            name: str = None,
            placeholder: str = "",
            spell_check: str = "true",
            min_length: str = "null",
            max_length: str = "null",
            pattern: str = "null",
            **kwargs
    ):
        return CharInput(
            *args, 
            name=name,
            placeholder=placeholder,
            spell_check=spell_check,
            type="email",
            min_length=min_length,
            max_length=max_length,
            pattern=pattern,
            **kwargs
        )


@dataclass(eq=False)
class SearchInput(HTMLElement):
    name: str
    placeholder: str
    spell_check: T.Union[str, None] = field(default=None)
    min_length: T.Union[str, None] = field(default=None)
    max_length: T.Union[str, None] = field(default=None)
    pattern: T.Union[str, None] = field(default=None)

    def __post_init__(self, *args, **kwargs):
        super(SearchInput, self).__init__(*args,
                                          name=self.name,
                                          placeholder=self.placeholder,
                                          spell_check=self.spell_check,
                                          min_length=self.min_length,
                                          max_length=self.max_length,
                                          pattern=self.pattern,
                                          **kwargs)

    def __render__(
            self,
            *args,
            name: str = None,
            placeholder: str = "",
            spell_check: str = "true",
            min_length: str = "null",
            max_length: str = "null",
            pattern: str = "null",
            **kwargs
    ):
        return CharInput(
            *args,
            name=name,
            placeholder=placeholder,
            type="search",
            spell_check=spell_check,
            min_length=min_length,
            max_length=max_length,
            pattern=pattern,
            **kwargs
        )


@dataclass(eq=False)
class CharField(HTMLElement):
    labeled = CharLabelValidator(logger=False, debug=True)
    input = CharInputValidator(logger=False, debug=True)

    label: str
    name: str
    placeholder: str
    type: str

    def __post_init__(self, *args, **kwargs):
        super(CharField, self).__init__(*args,
                                        label=self.label,
                                        name=self.name,
                                        placeholder=self.placeholder,
                                        type=self.type,
                                        **kwargs
                                        )

    def __render__(self, *args, label, name, placeholder, type, **kwargs):
        self.labeled = CharLabel(label)
        self.input = CharInput(
                  name=name,
                  placeholder=placeholder,
                  type=type
              )
        return self.html_tags.div(*args, self.labeled, self.input, **kwargs)


class CharFieldSet(HTMLElement):

    def __render__(self, legend: CharLegend, *fields: HTMLElement, **kwargs):
        return self.html_tags.fieldset(legend, *fields, **kwargs)


if __name__ == '__main__':
    print(CharField(label="Cut", name="cut", placeholder="Diamond Cut", type="text"))
    print(HiddenInput(name="password", placeholder="Enter Password", min_length="4"))
    print(CharInput(name="text", placeholder="Enter Password", min_length="13", spell_check="false"))
    


# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


import typing as T
from enum import Enum, IntEnum

from uidom.dom.htmlelement import HTMLElement
from uidom.dom.src.htmltags import option, select
from valio import (EnumValidator, IntegerEnumValidator, StringEnumValidator,
                   StringValidator)

__all__ = [
    "EnumLabel",
    "EnumInput",
    "IntegerEnumInput",
    "CharEnumInput"
]


class EnumLabel(HTMLElement):
    label = StringValidator(logger=False, debug=True)

    def __render__(self, label, *args, **kwargs):
        self.label = label
        return self.html_tags.label(self.label, *args, **kwargs)


class EnumInput(HTMLElement):
    options: Enum = EnumValidator(logger=False, debug=True)

    def __render__(self, *args, options, **kwargs):
        self.options = options
        return select(option(opt.value, value=str(opt.value).capitalize())
                      if len(opt.value) == 1 else option(opt.value[0], value=str(opt.value[1]))
                      for opt in options)


class IntegerEnumInput(HTMLElement):
    options: IntEnum = IntegerEnumValidator(logger=False, debug=True)

    def __render__(self, *args, options, **kwargs):
        self.options = options
        return select(option(opt.value, value=str(opt.value).capitalize())
                      if len(opt.value) == 1 else option(opt.value[0], value=str(opt.value[1]))
                      for opt in options)


class CharEnumInput(HTMLElement):
    options: T.Union[str, Enum, None] = StringEnumValidator(logger=False, debug=True)

    def __render__(self, *args, options, **kwargs):
        self.options = options
        return select(option(opt.value, value=str(opt.value).capitalize())
                      if len(opt.value) == 1 else option(opt.value[0], value=str(opt.value[1]))
                      for opt in options)




# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT



from dataclasses import asdict, dataclass, field
from typing import Iterable, Union

from uidom.dom.src import csstags, htmltags, jinjatags, svgtags, vuetifytags
from uidom.dom.src.main import extension

__all__ = [
    "Component"
]


@dataclass
class Component(extension.Tags):
    left_delimiter = "<"
    right_delimiter = ">"
    css_tags = csstags
    svg_tags = svgtags
    html_tags = htmltags
    jinja_tags = jinjatags
    vuetify_tags = vuetifytags
    file_extension:str = field(init=False, default='')
    render_tag: bool = field(init=False, default=False)
    children: list = field(init=False, default_factory=list)
    document: Union[extension.Tags, None] = field(init=False, default=None)
    attributes: dict = field(init=False, default_factory=dict)

    def __init__(self, *args, **kwargs):
        super(Component, self).__init__()
        self._entry = super(Component, self).add(self.__render__(*args, **kwargs))
        # we perform checks on the _entry "after" the dom initialization because .get method
        # looks into children
        self.__checks__(self._entry)

    def __checks__(self, element: extension.Tags) -> extension.Tags:  # noqa
        if self.render_tag:
            raise ValueError(f"{self.render_tag} can not be true for Components")
        return element

    def add(self, *args):
        '''
        Adding tags to a component appends them to the __render__.
        '''
        return self._entry.add(*args)

    def set_attribute(self, key, value):
        if not self.render_tag:
            self.children[0].set_attribute(key, value)
        else:
            super(Component, self).set_attribute(key, value)

    __setitem__ = set_attribute

    def delete_attribute(self, key):
        if not self.render_tag:
            self.children[0].delete_attribute(key)
        else:
            super(Component, self).delete_attribute(key)

    __delitem__ = delete_attribute

    def get(self, tag=None, **kwargs):
        if not self.render_tag:
            return self.children[0].get(tag, **kwargs)
        else:
            return super(Component, self).get(tag, **kwargs)

    def __getitem__(self, key):
        if not self.render_tag:
            if any(object.__getattribute__(self, 'children')):
                return object.__getattribute__(self, 'children')[0].__getitem__(key)
        return super(Component, self).__getitem__(key)

    __getattr__ = __getitem__
    
    def asdict(self, exclude=None):
        exclude = exclude or ['file_extension', 'render_tag', 'children', 'document', 'attributes']
        return {key: value for key, value in asdict(self).items() if key not in exclude} 

    def __hash__(self) -> int:
        return hash(self._entry)

    def __render__(self, *args, **kwargs) -> htmltags.html_tag:  # noqa
        raise NotImplementedError(f"method: {self.__render__.__qualname__} not implemented")

    def call(self, *args, **kwargs):
        """
        This method is basically placeholder for using websocket communications.
        All sorts of fun stuffs can happens here.
        :param args:
        :param kwargs:
        :return:
        """

        raise NotImplemented(f"method: {self.call.__qualname__} not implemented")

    def __dir__(self) -> Iterable[str]:
        return sorted(iter(self.__dict__), key=lambda k: k)


if __name__ == '__main__':
    from valio import StringValidator

    # using @dataclass(eq=False) to use super class hash function
    @dataclass(eq=False)
    class vue(Component):
        a: str = StringValidator(logger=False, debug=True)
        
        def __post_init__(self):
            super(vue, self).__init__(a=self.a)
            
        def __render__(self, a) -> htmltags.html_tag: # type: ignore[override]
            return self.html_tags.p(a=a)

    print(vue(a='1'))



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


import typing
from dataclasses import dataclass

from uidom.components.alpinejs.slot_element import x_slot
from uidom.dom.htmlelement import *
from uidom.dom.src.htmltags import *

__all__ = [
    "WebComponentSlot"
]


@dataclass
class WebComponentSlot(XComponent):
    slot_names: typing.Union[list, tuple]
    classes: dict
    css: list[str]
    slot_class: str = ''

    def __post_init__(self):
        super(WebComponentSlot, self).__post_init__(
            slot_names=self.slot_names,
            classes=self.classes,
            css=self.css,
            slot_class=self.slot_class
        )

    def __render__(
            self,
            tag_name,
            slot_names,
            classes,
            css,
            slot_class
    ):
        return div(
            template(
                x_slot(
                    slotnames=slot_names,
                    container_part=f"{tag_name}",
                    classes=classes,
                    css=css,
                    exportparts="*",
                    className=slot_class if slot_class != '' else False,
                ), tabindex=0
            ),
            x_component=tag_name, shadowroot="true"
        )

    def __checks__(self, element):
        # XComponent.__checks__(self, element)
        # return self.__shadowroot_checks(element)
        return element

    def __call__(self, *args, **kwargs):
        return super(WebComponentSlot, self).__call__(*args, exportparts="*", **kwargs)



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.htmlelement import HTMLElement
from uidom.dom.src import a, div

__all__ = [
    "Link",
    "HtmxLink"
]


class Link(HTMLElement):

    def __render__(self, *args, item_link, item_name=None, **kwargs):
        self.div = div() if item_name is None else div(item_name)
        return a(*args, self.div, href=item_link,  **kwargs)


class HtmxLink(HTMLElement):

    def __render__(self, *args, item_link, target_id, item_name=None, push_url=True, **kwargs):
        self.text = div() if item_name is None else div(item_name)
        link = div(*args, self.text, hx_get=item_link, hx_trigger="click", hx_target=target_id,
                   className="cursor-pointer",
                   **kwargs)
        if push_url:
            link["hx-push-url"] = "true"
        return link



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


import socketserver
from dataclasses import dataclass
from http.server import SimpleHTTPRequestHandler

from uidom.dom import *

PORT = 8000

Handler = SimpleHTTPRequestHandler


@dataclass
class InitializationServer(HTMLElement):

    def __render__(self, *args, **kwargs):
        ...

    def run_server(self, ):
        with socketserver.TCPServer(("", PORT), Handler) as httpd:
            print("serving at port", PORT)
            httpd.serve_forever()



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom import *

__all__ = [
    "TogglePlain",
    "ToggleOutset",
    "ToggleInset"
]


class TogglePlain(AlpineComponent):
    # from https://alpineapple.dev/toggle/
    x_data = {"toggle": "false"}

    def __render__(self, tag_name):
        # "{{'toggle': 'false'}}" <--- how single-quotes ['false', 'toggle'] etc are used
        # to correctly parse f-string
        return HTMLStringToDom(f'''
        <div x-data="{{'toggle': 'false'}}" 
            @click="toggle = !toggle" 
            x_component="{tag_name}"
            class="relative cursor-pointer h-6 w-12 rounded-full transition-all duration-300 
                   ease-in-out bg-cyan-400 bg-slate-400" :class=" toggle ? 'bg-cyan-400' : 'bg-slate-400' ">
            <span class="absolute h-6 w-6 rounded-full bg-white transition-all duration-300 ease-in-out" :class=" toggle ? 'translate-x-7' : '' "></span>
        </div>
        ''').parse()


class ToggleInset(XComponent):

    def __render__(self, tag_name):
        return HTMLStringToDom(f'''
        <template x-component="{tag_name}" >
            <label for="Toggle1" x-data="$el.parentElement.data()" class="inline-flex items-center space-x-4 cursor-pointer dark:text-gray-100">
                <span>Left</span>
                <span class="relative">
                    <input id="Toggle1" type="checkbox" class="hidden peer">
                    <div class="w-10 h-6 rounded-full shadow-inner dark:bg-gray-400 peer-checked:dark:bg-violet-400"></div>
                    <div class="absolute inset-y-0 left-0 w-4 h-4 m-1 rounded-full shadow peer-checked:right-0 peer-checked:left-auto dark:bg-gray-800"></div>
                </span>
                <span>Right</span>
            </label>
        </template>
        ''').parse()


class ToggleOutset(XComponent):

    def __render__(self, tag_name):
        return HTMLStringToDom(f'''
        <template x-component={tag_name}>
            <label for="Toggle1" class="inline-flex items-center space-x-4 cursor-pointer dark:text-gray-100">
                <span>Left</span>
                <span class="relative">
                    <input id="Toggle1" type="checkbox" class="hidden peer">
                    <div class="w-10 h-6 rounded-full shadow-inner dark:bg-gray-400 peer-checked:dark:bg-violet-400"></div>
                    <div class="absolute inset-y-0 left-0 w-4 h-4 m-1 rounded-full shadow peer-checked:right-0 peer-checked:left-auto dark:bg-gray-800"></div>
                </span>
                <span>Right</span>
            </label>
        </template>
        ''').parse()


if __name__ == '__main__':
    print(TogglePlain('xyz'))



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from dataclasses import dataclass, field

from uidom.dom.htmlelement import *
from uidom.dom.src.htmltags import *

__all__ = [
    "SlotElement",
    "x_slot"
]


@dataclass
class SlotElement(CustomElement):
    # fields = [
    #     ("slotnames", list[str], field(default_factory=lambda: [''])),
    #     ("classes", dict[str, str], field(default_factory=lambda:{})),
    #     ("css", list[str], field(default_factory=lambda:[''])),
    #     ("container_part", str, field(default='')),
    #     ("exportparts", str, field(default=''))
    # ]

    def __render__(self, tag_name):
        """
        :param tag_name:
        :return:
        """
        return template(
            div(
                template(link(x_bind_href="ss", rel="stylesheet", type="text/css"), x_for="ss in css"),
                template(
                    div(slot(x_bind_name="name"), x_bind_part="name", x_bind_class="classes[name]",
                        exportparts="*"),
                    x_for="name in slotnames", tabindex=0,
                    x_if=" !!slotnames && slotnames.length && slotnames[0] !== '' "
                ),
                template(
                    slot(),
                    tabindex=0,
                    x_if=" !slotnames || !slotnames.length || (slotnames.length && slotnames[0] == '') ",
                ),
                x_bind_class="classes[`${container_part}`]",
                x_bind_part="container_part",
                exportparts="*",
                x_data="{...slots(), ...$el.parentElement.data()}", x_cloak=None,
            ),
            script(
                '''function slots(){
                return {
                    slotnames: [''],
                    classes:{}, 
                    css:[''] , 
                    container_part: ''
                        }
                    }'''
            ),
            x_component=tag_name,
        )


x_slot = SlotElement(tag_name="slot")

if __name__ == '__main__':
    print(x_slot())



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.htmlelement import HTMLElement
from uidom.dom.src.htmltags import script
from uidom.dom.src.utils.dom_util import raw

# The pure client-side code taken from https://stackoverflow.com/a/67276585
# and https://stackoverflow.com/a/48777481


class XShadowComponent(HTMLElement):

    # "init-alpine-tree" directive signal's browser to initialise the alpine tree with shadow-dom

    def __render__(self, *args, **kwargs):
        return script(raw("""
    document.querySelectorAll('[x-shadow-component]').forEach(component => {
        const componentName = `x-${component.getAttribute('x-shadow-component')}`
        class XShadowComponent extends HTMLElement {
            constructor() {
                super();
                var shadow = this.attachShadow({mode: 'open'});
                var slot = document.createElement('slot');
                var template = component.getElementsByTagName('template');
                if (template.length > 0) {
                    shadow.appendChild(template[0].content.cloneNode(true));
                    }
                else {
                    shadow.appendChild(slot);
                    }
                this.shadow = shadow
                }
                        
            connectedCallback() {
               if (component.hasAttribute("init-alpine-tree") && component.getAttribute("init-alpine-tree")) {
                    document.addEventListener('alpine:initialized', () => Alpine.initTree(this.shadow));
                }
            }

            attributeChangedCallback() {
                super.attributeChangedCallback();
            }

            disconnectedCallback() {
                super.disconnectedCallback();
            }

            data() {
                const attributes = this.getAttributeNames()
                const data = {}

                function isJSON(str) {
                    try {
                        return (JSON.parse(str) && !!str);
                    } catch (e) {
                        return false;
                        }
                    }


                attributes.forEach(attribute => {

                    if (!isJSON(this.getAttribute(attribute))) {
                        data[attribute] = this.getAttribute(attribute)
                    } else {
                        data[attribute] = JSON.parse(this.getAttribute(attribute))   
                    }
                })
                return data
            }
        }
        
        customElements.define(componentName, XShadowComponent)
    })
"""))



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

# Copyright (c) 2022 Valio
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

from uidom.dom.src.main import extension


class VuetifyTags(extension.DoubleTags):
    child_dedent = False
    self_dedent = False
    left_tag = "<"
    right_tag = ">"


class VApp(VuetifyTags):
    tagname = "v-app"


class VAppBar(VuetifyTags):
    tagname = "v-app-bar"


class VAppBarNavIcon(VuetifyTags):
    tagname = "v-app-bar-nav-icon"


class VToolBarTitle(VuetifyTags):
    tagname = "v-toolbar-title"


class VToolbar(VuetifyTags):
    tagname = "v-toolbar"


class VSpacer(VuetifyTags):
    tagname = "v-spacer"


class VMain(VuetifyTags):
    tagname = "v-main"


class VContainer(VuetifyTags):
    tagname = "v-container"


class component(VuetifyTags):
    pass


if __name__ == '__main__':

    from uidom.dom.htmlelement import HTMLElement
    from uidom.dom.src.htmltags import comment

    class x(HTMLElement):

        def __render__(self, name, *args, **kwargs):
            return component(name, *args, **kwargs)

    from uidom.dom.src.htmltags import body, script
    print(x(body(script("jajajajaj"), comment("hahahah")), hello=1, x_dot_data="jk", _="sa"))



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.htmlelement import HTMLElement


class Flex(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " flex"
        return elem


class Flex1(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " flex-1"
        return elem


class FlexAuto(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " flex-auto"
        return elem


class FlexInitial(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " flex-initial"
        return elem


class FlexNone(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " flex-none"
        return elem




# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.src import component

__all__ = [
    "JinjaComponent"
]


class JinjaComponent(component.Component):
    file_extension = ".html"

    def __render__(self, *args, **kwargs):
        pass

    def __block__(self, block_name, *args):
        return self.jinja_tags.Block(block_name, *args)

    def __if__(self, if_condition, *if_args):
        return self.jinja_tags.If(if_condition, *if_args)

    def __elif__(self, elif_condition, *elif_args):
        return self.jinja_tags.Elif(elif_condition, *elif_args)

    def __else__(self, *else_args):
        return self.jinja_tags.Else(*else_args)

    def __if_else__(self, if_condition, *else_args):
        return self.__if__(if_condition, self.__else__(*else_args))

    def __if_elif__(self, if_condition, elif_condition, *elif_args):
        self.__if__(if_condition, self.__elif__(elif_condition, *elif_args))

    def __load__(self, load, *args):
        return self.jinja_tags.Load(load, *args)

    def __include__(self, include, *args):
        return self.jinja_tags.Include(include, *args)

if __name__ == '__main__':
    from uidom.dom import Elif, Else, If, p

    print(If("test", p("testing", p('ok')), Elif("staging", p("staging")), Else(p("production"), p("a"))))



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

import os
import sys
import typing as T
from dataclasses import dataclass
from pathlib import Path

from uidom.dom.htmldocument import HtmlDocument
from uidom.dom.src import ext
from uidom.settings.paths import make_paths

__all__ = ["UiDOM", "WebAssets", "FileSettings"]


@dataclass
class UiDOM(object):
    head: T.Optional[T.Union[ext.Tags, list[ext.Tags]]] = None
    body: T.Optional[T.Union[ext.Tags, list[ext.Tags]]] = None
    ensure_csrf_token_in_meta: T.Union[bool, None] = False
    webassets: T.Optional["WebAssets"] = None

    def __call__(self, *args, head=None, body=None, **kwargs) -> HtmlDocument:
        html_doc = HtmlDocument
        if not self.ensure_csrf_token_in_meta:
            html_doc.ensure_csrf_token_in_meta = self.ensure_csrf_token_in_meta
        return html_doc(
            *args,
            head=head,
            body=body,
            common_head=self.head,
            common_body=self.body,
            **kwargs
        )


@dataclass
class FileSettings(object):
    BASE_DIR: T.Union[str, Path]
    SUB_DIR: T.Union[str, Path] = ""
    TEMPLATE_PATH = "template"
    CACHE_PATH = "cache"
    STATIC_PATH = "static"
    DATABASE_PATH = "database"
    UPLOAD_PATH = "upload"
    MEDIA_PATH = "media"
    FILE_PATH = "file"
    CSS_PATH = "css"
    JS_PATH = "js"
    FONT_PATH = "font"
    IMAGE_PATH = "image"
    AUDIO_PATH = "audio"
    VIDEO_PATH = "video"

    SQL_PATH = "sql"
    NOSQL_PATH = "no_sql"

    def __post_init__(self):
        # if not all([self.BASE_DIR, self.SUB_DIR]):
        #     raise ValueError(f"BASE_DIR, BASE_PATH values required")
        if isinstance(self.BASE_DIR, str):
            if os.path.isfile(self.BASE_DIR):
                self.BASE_DIR = Path(self.BASE_DIR).parent.resolve()
            else:
                self.BASE_DIR = Path(self.BASE_DIR).resolve()


class DirConfig(object):

    def __init__(self, file_settings: FileSettings):
        subdir = (
            file_settings.SUB_DIR
            if not os.path.isfile(file_settings.SUB_DIR)
            else os.path.dirname(file_settings.SUB_DIR)
        )

        root_dir = (
            os.path.join(file_settings.BASE_DIR, subdir)
            if os.path.isdir(file_settings.BASE_DIR)
            else os.path.join(file_settings.BASE_DIR.replace(".", os.path.sep), subdir)
            if not os.path.isfile(file_settings.BASE_DIR)
            else os.path.join(os.path.dirname(file_settings.BASE_DIR), subdir)
            if file_settings.BASE_DIR != "__main__"
            else os.path.join(os.path.dirname(sys.modules["__main__"].__file__), subdir)
        ) if isinstance(file_settings.BASE_DIR, str) \
            else os.path.join(file_settings.BASE_DIR, subdir) if any(subdir) else file_settings.BASE_DIR

        self.ROOT_DIR = Path(
            root_dir if any([root_dir]) else os.path.abspath(f".{os.path.sep}")
        )
        self.TEMPLATE_DIR = self.ROOT_DIR / file_settings.TEMPLATE_PATH
        self.CACHE_DIR = self.ROOT_DIR / file_settings.CACHE_PATH
        self.DATABASE_DIR = self.ROOT_DIR / file_settings.DATABASE_PATH
        self.STATIC_DIR = self.ROOT_DIR / file_settings.STATIC_PATH
        self.UPLOAD_DIR = self.ROOT_DIR / file_settings.UPLOAD_PATH

        self.STATIC_MEDIA_DIR = self.STATIC_DIR / file_settings.MEDIA_PATH
        self.STATIC_FILE_DIR = self.STATIC_DIR / file_settings.FILE_PATH

        self.STATIC_CSS_DIR = self.STATIC_FILE_DIR / file_settings.CSS_PATH
        self.STATIC_JS_DIR = self.STATIC_FILE_DIR / file_settings.JS_PATH
        self.STATIC_FONT_DIR = self.STATIC_FILE_DIR / file_settings.FONT_PATH

        self.STATIC_IMAGE_DIR = self.STATIC_MEDIA_DIR / file_settings.IMAGE_PATH
        self.STATIC_AUDIO_DIR = self.STATIC_MEDIA_DIR / file_settings.AUDIO_PATH
        self.STATIC_VIDEO_DIR = self.STATIC_MEDIA_DIR / file_settings.VIDEO_PATH

        self.UPLOAD_MEDIA_DIR = self.UPLOAD_DIR / file_settings.MEDIA_PATH
        self.UPLOAD_FILE_DIR = self.UPLOAD_DIR / file_settings.FILE_PATH
        self.UPLOAD_IMAGE_DIR = self.UPLOAD_MEDIA_DIR / file_settings.IMAGE_PATH
        self.UPLOAD_AUDIO_DIR = self.UPLOAD_MEDIA_DIR / file_settings.AUDIO_PATH
        self.UPLOAD_VIDEO_DIR = self.UPLOAD_MEDIA_DIR / file_settings.VIDEO_PATH

        self.SQL_DATABASE_DIR = self.DATABASE_DIR / file_settings.SQL_PATH
        self.NOSQL_DATABASE_DIR = self.DATABASE_DIR / file_settings.NOSQL_PATH

    def __iter__(self):
        for item in self.__dict__:
            if "ROOT" in item or "DIR" in item:
                yield self.__dict__[item]

    def __next__(self):
        return next(self)

    def dict(self):
        return {item: self.__dict__[item]
                for item in self.__dict__
                if "ROOT" or "DIR" in item}

    def __str__(self):
        return str(list(self))


class UrlConfig(object):
    def __init__(self, dir_config: DirConfig):
        self.ROOT_URL = dir_config.ROOT_DIR
        self.TEMPLATE_URL = Path(os.path.sep) / dir_config.TEMPLATE_DIR.relative_to(dir_config.ROOT_DIR)
        self.STATIC_URL = Path(os.path.sep) / dir_config.STATIC_DIR.relative_to(dir_config.ROOT_DIR)
        self.UPLOAD_URL = Path(os.path.sep) / dir_config.UPLOAD_DIR.relative_to(dir_config.ROOT_DIR)

        self.STATIC_MEDIA_URL = Path(os.path.sep) / dir_config.STATIC_MEDIA_DIR.relative_to(dir_config.STATIC_DIR)
        self.STATIC_FILE_URL = Path(os.path.sep) / dir_config.STATIC_FILE_DIR.relative_to(dir_config.STATIC_DIR)

        self.STATIC_CSS_URL = Path(os.path.sep) / dir_config.STATIC_CSS_DIR.relative_to(dir_config.STATIC_FILE_DIR)
        self.STATIC_JS_URL = Path(os.path.sep) / dir_config.STATIC_JS_DIR.relative_to(dir_config.STATIC_FILE_DIR)
        self.STATIC_FONT_URL = Path(os.path.sep) / dir_config.STATIC_FONT_DIR.relative_to(dir_config.STATIC_FILE_DIR)
        self.STATIC_IMAGE_URL = Path(os.path.sep) / dir_config.STATIC_IMAGE_DIR.relative_to(dir_config.STATIC_MEDIA_DIR)
        self.STATIC_AUDIO_URL = Path(os.path.sep) / dir_config.STATIC_AUDIO_DIR.relative_to(dir_config.STATIC_MEDIA_DIR)
        self.STATIC_VIDEO_URL = Path(os.path.sep) / dir_config.STATIC_VIDEO_DIR.relative_to(dir_config.STATIC_MEDIA_DIR)

        self.UPLOAD_MEDIA_URL = Path(os.path.sep) / dir_config.UPLOAD_MEDIA_DIR.relative_to(dir_config.UPLOAD_DIR)
        self.UPLOAD_IMAGE_URL = Path(os.path.sep) / dir_config.UPLOAD_IMAGE_DIR.relative_to(dir_config.UPLOAD_MEDIA_DIR)
        self.UPLOAD_AUDIO_URL = Path(os.path.sep) / dir_config.UPLOAD_AUDIO_DIR.relative_to(dir_config.UPLOAD_MEDIA_DIR)
        self.UPLOAD_VIDEO_URL = Path(os.path.sep) / dir_config.UPLOAD_VIDEO_DIR.relative_to(dir_config.UPLOAD_MEDIA_DIR)
        self.UPLOAD_FILE_URL = Path(os.path.sep) / dir_config.UPLOAD_FILE_DIR.relative_to(dir_config.UPLOAD_DIR)

        self.SQL_DATABASE_URL = Path(os.path.sep) / dir_config.SQL_DATABASE_DIR.relative_to(dir_config.ROOT_DIR)

    def __iter__(self):
        for item in self.__dict__:
            if "URL" in item:
                yield self.__dict__[item]

    def __next__(self):
        return next(self)

    def dict(self):
        return {item: self.__dict__[item]
                for item in self.__dict__
                if "URL" in item}

    def __str__(self):
        return str(list(self))


class DatabaseFile(object):

    def __init__(self, dir_config: DirConfig):
        self.sql: Path = dir_config.SQL_DATABASE_DIR
        self.nosql: Path = dir_config.NOSQL_DATABASE_DIR


class StaticFile(object):
    def __init__(self, dir_config: DirConfig):
        self.dir: Path = dir_config.STATIC_DIR
        self.images: Path = dir_config.STATIC_IMAGE_DIR
        self.audios: Path = dir_config.STATIC_AUDIO_DIR
        self.videos: Path = dir_config.STATIC_VIDEO_DIR
        self.files: Path = dir_config.STATIC_FILE_DIR
        self.css: Path = dir_config.STATIC_CSS_DIR
        self.js: Path = dir_config.STATIC_JS_DIR
        self.fonts: Path = dir_config.STATIC_FONT_DIR
        self.medias: Path = dir_config.STATIC_MEDIA_DIR


class UploadedFile(object):
    def __init__(self, dir_config: DirConfig):
        self.dir: Path = dir_config.UPLOAD_DIR
        self.images: Path = dir_config.UPLOAD_IMAGE_DIR
        self.audios: Path = dir_config.UPLOAD_AUDIO_DIR
        self.videos: Path = dir_config.UPLOAD_VIDEO_DIR
        self.files: Path = dir_config.UPLOAD_FILE_DIR
        self.medias: Path = dir_config.UPLOAD_MEDIA_DIR


class TemplateFile(object):
    def __init__(self, dir_config: DirConfig):
        self.dir: Path = dir_config.TEMPLATE_DIR


class WebAssets(object):
    files: FileSettings
    directory: DirConfig
    url: UrlConfig
    template: TemplateFile
    upload: UploadedFile
    static: StaticFile
    database: DatabaseFile

    def __init__(self, file_settings: FileSettings):

        # basic settings
        self.directory = DirConfig(file_settings=file_settings)
        self.url = UrlConfig(dir_config=self.directory)

        # files
        self.template = TemplateFile(dir_config=self.directory)
        self.upload = UploadedFile(dir_config=self.directory)
        self.static = StaticFile(dir_config=self.directory)
        self.database = DatabaseFile(dir_config=self.directory)
        files_created = make_paths(self.directory)
        if files_created:
            print("files created :: ")
            for file in files_created:
                print("\tfile: ", file)


if __name__ == '__main__':
    from pathlib import Path

    from uidom.dom import div, meta, title
    doc = UiDOM()
    print(doc("hahah"))



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.htmlelement import HTMLElement
from uidom.dom.src.htmltags import script
from uidom.dom.src.utils.dom_util import raw

# The pure client-side code taken from https://stackoverflow.com/a/67276585
# and https://stackoverflow.com/a/48777481

__all__ = ["XComponentJS"]


class XComponentJS(HTMLElement):
    """
    Adaptor concept taken from:
        https://github.com/material-components/material-components-web/tree/9736ddce9c12f5485e746984225919568541e88d/packages/mdc-base
    """

    def __render__(self, *args, **kwargs):
        js = """
            
            function guidGenerator() {
                const S4 = function (){
                    return (((1 + Math.random()) * 0x10000) | 0).toString(16).substring(1);
                };
                return S4() + S4() + '-' + S4() + '-' + S4() + '-' + S4() + '-' + S4() + S4() + S4();
            }
        
            const toCamel = (s) => {
                return s.replace(/([-_][a-z])/ig, ($1) => {
                    return $1.toUpperCase()
                        .replace('-', '')
                        .replace('_', '');
                });
            }

            const isJSON = (str) => {
                try {
                    return (JSON.parse(str) && !!str);
                } catch (e) {
                    return false;
                }
            }

            const observeAttrChange = (el, attributeChangedCallback) => {
                var observer = new MutationObserver(mutations => {
                    mutations.forEach((mutation) => {
                        if (mutation.type === 'attributes') {
                            let oldVal = mutation.oldValue;
                            let newVal = mutation.target.getAttribute(mutation.attributeName);
                            attributeChangedCallback(mutation.attributeName, oldVal, newVal);
                        }
                    });
                });
                observer.observe(el, {
                    attributes: true,
                    attributeOldValue: true,
                    attributeFilter: ['id', 'class', 'style']
                });
                return observer;
            }

            (() => {
                let socket = {};
                let messageHandler = {};
                
                const setUpOrGetWebSocket = (elementID, messageHandler, endPoint = 'ws') => {
                    if (!socket[`${endPoint}`]) {
                        let url = new URL(document.location);
                        const wsUrl = `${url.protocol.replace('http', 'ws')}//${url.hostname}:${url.port}/${endPoint}`;
                        socket[`${endPoint}`] = new WebSocket(wsUrl);
                        socket[`${endPoint}`].onopen = (event) => {
                            console.log(`connected to ${wsUrl}`);
                            socket[`${endPoint}`].send(`connection to client user ${elementID}:${JSON.stringify(event)}`);
                            console.log(event);
                        };
                        socket[`${endPoint}`].onclose = (event) => {
                            console.log(`connection closed to ${elementID}:${wsUrl}`);
                            socket[`${endPoint}`] = null;
                            setTimeout(
                                () => {socket[`${endPoint}`] = setUpOrGetWebSocket(elementID, messageHandler, endPoint)}
                                , 300);
                            clearTimeout();
                        };
                        socket[`${endPoint}`].onerror = (event) => {
                            console.log(`connection closed to ${elementID}:${wsUrl} due to error`);
                        };
                    }
                    socket[`${endPoint}`].onmessage = (message) => {
                        let data = message.data;
                        if (isJSON(data)) {
                            data = JSON.parse(data);
                        }
                        messageHandler[`${elementID}`](data);
                    };
                    return socket[`${endPoint}`];
                };
                document.setUpOrGetWebSocket = setUpOrGetWebSocket;
                document.messageHandler = messageHandler;
            })();

            
            document.querySelectorAll('[x-component]').forEach(component => {
                const componentName = `x-${component.getAttribute('x-component')}`;

                class XComponent extends HTMLElement {
                    
                    constructor() {
                        super();
                    }

                    connectedCallback() {
                        var isShadowRoot = component.getAttribute('shadowroot');
                        if (!!isShadowRoot){
                            
                            let shadow = this.attachShadow({mode: 'open'});
                            
                            let template;
                            if (component.tagName === 'TEMPLATE'){
                                template = component;
                            } else {
                                template = component.getElementsByTagName('template')[0];
                            }

                            if (template?.content.childElementCount) {
                                shadow.appendChild(template.content.cloneNode(true));
                                }
                            else {
                                let slot = document.createElement('slot');
                                shadow.appendChild(slot);
                                }
                            this.shadow = shadow;
                            this.checkId();
                            document.addEventListener('alpine:initialized', () => {
                                Alpine.initTree(this.shadow);
                            });
                        }
                        
                        if (!this.id){
                            this.append(component.content.cloneNode(true));
                            this.checkId();
                        }
                        
                        this.addEventListener('alpine:init', () => {
                            Alpine.data("parentElementData", () => ({
                                ...this.data()
                            }))
                        });
                        
                              
                        this.dataCallback = this.dataCallback.bind(this);              
                        // <my-div active="true", data-target="#accordian"></my-div>
                        // example: for setting up  name="cool-dude" value as this.name value
                        let _dataState = this.data();

                        // for setting up data-target value
                        let _dataTarget = this?.dataset.target || {};
                        if (isJSON(_dataTarget)){
                            _dataTarget = JSON.parse(_dataTarget);
                        }
                        // for setting up data-action value
                        let _dataAction = this?.dataset.action || {};
                        //console.log("DATA_STATE: ", _dataState);
                        if (!!_dataState) {
                            Object.keys(_dataState).forEach(attr => {
                                if (this.hasOwnProperty(attr)){
                                    throw new Error(`attribute ${attr} can't be assigned as its builtin HtmlElement property`);
                                }
                                Object.defineProperty(this, `_data_${attr}`, {
                                    get() {
                                        if (this.hasAttribute(`__${attr}`)) {
                                            return this.getAttribute(`__${attr}`);
                                        } else {
                                            let val = _dataState[attr];
                                            if (isJSON(val)) {
                                                this.setAttribute(`__${attr}`, JSON.parse(val));
                                            } else {
                                                this.setAttribute(`__${attr}`, val);
                                            }
                                            return this.getAttribute(`__${attr}`);
                                        }
                                    },
                                    set(value) {
                                        if (isJSON(value)) {
                                            this.setAttribute(`__${attr}`, JSON.parse(value));
                                        } else {
                                            this.setAttribute(`__${attr}`, value);
                                        }
                                        let oldValue = this.dataset.state[attr];
                                        if (oldValue !== value) {
                                            this.dataset.state[attr] = value;
                                        }
                                    }
                                });
                            });
                        }

                        // connecting to websockets
                        if (_dataState?.ws){
                            let messageHandler = document.messageHandler;
                            messageHandler[`${this.id}`] = () => {};
                            this.ws = document.setUpOrGetWebSocket(this.id, messageHandler, _dataState?.ws || 'ws');
                        }
                        if  (!!this?.ws?.readyState){
                            this.ws.send(JSON.stringify({id: this.id}));
                        }
                        
                        
                        this.observer = observeAttrChange(this, (attr, oldVal, newVal) => {
                            // slice :attr to remove leading '_' (underscore) to check in _dataState
                            //console.log(attr, _dataState.includes(attr.replace('__', '')));
                            if (_dataState?.includes(attr.replace('__', ''))) {
                                this.attributeChangedCallback(attr.replace('__', ''), oldVal, newVal);
                            }
                        });
                    }
                    
                    checkId() {
                        if (!this.id) this.id = `x-${component.getAttribute('x-component')}` + '-' + guidGenerator();
                    }
                    
                    dataCallback(message){
                        if (isJSON(message)){
                            message = JSON.parse(message);
                        }
                        if (typeof message === "[object String]"){
                            console.log("SERVER_MESSAGE", message);
                            return;
                        }
                        if (this.id === message.id){
                            this.attributeChangedCallback(message.attr, message.oldVal, message.newVal);
                        }
                        this.ws.send(
                            JSON.stringify(
                                {id: message.id , 
                                 attr:message.attr,
                                 oldVal: message.oldVal, 
                                 newVal: message.newVal}
                            )
                        );
                    }

                    attributeChangedCallback(attrName, o, n) {
                        console.log(attrName, o, n);
                        if (n !== o){
                            this[attrName] = n;   
                            //this.dataCallback(attrName, o, n);
                        }
                    }

                    disconnectedCallback() {
                        //super.disconnectedCallback();
                        this.observer.disconnect();
                        this.observer = null;
                    }

                    async getFile(url) {
                        const res = await fetch(url);
                        if (!res.ok) {
                            throw Object.assign(new Error(res.statusText + ' ' + url), {
                                res
                            });
                        }
                        return {
                            getContentData: asBinary => asBinary ? res.arrayBuffer() : res.text(),
                        };
                    }


                    dispatch(name, data, options = {
                        bubble: true,
                        cancelable: false,
                        composed: false
                    }) {
                        const event = new CustomEvent(name, {
                            bubbles: options.bubble,
                            cancelable: options.cancelable,
                            composed: options.composed,
                            detail: data
                        });
                        this.dispatchEvent(event);
                    }
                    

                    data() {
                        const attributes = this.getAttributeNames();
                        const data = {};

                        attributes.forEach(attribute => {

                            if (!isJSON(this.getAttribute(attribute))) {
                                data[attribute] = this.getAttribute(attribute);
                            } else {
                                data[attribute] = JSON.parse(this.getAttribute(attribute));
                            }
                        });
                        return data;
                    }

                }

                customElements.define(componentName, XComponent);
            }); 
            
            """
        return script(raw(js))



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.htmlelement import HTMLElement
from uidom.dom.tailwindcss import (Columns, Flex, FlexColumn, FlexRow,
                                        Grid, Rows)


class GridRowsColumnsDiv(HTMLElement):

    def __render__(self, *args, columns, rows, **kwargs):
        elem = self.html_tags.div(*args, **kwargs)
        elem["class"] = ""
        return Rows(Columns(Grid(elem), columns=columns), rows=rows)


class FlexColumnDiv(HTMLElement):

    def __render__(self, *args, **kwargs):
        elem = self.html_tags.div(*args, **kwargs)
        elem["class"] = ""
        return Flex(FlexColumn(elem))


class FlexRowDiv(HTMLElement):

    def __render__(self, *args, **kwargs):
        elem = self.html_tags.div(*args, **kwargs)
        elem["class"] = ""
        return FlexRow(Flex(elem))


if __name__ == '__main__':
    print(FlexRowDiv("hello world"))



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.elements import (CharFieldSet, CharLabel, CharLegend,
                                       PasswordInput, SubmitButton, TextInput)
from uidom.dom.htmlelement import HTMLElement
from uidom.dom.src.htmltags import div, form
from uidom.dom.src.svgtags import svg
from valio import Validator

__all__ = [
    "LoginForm",
    "FieldSetLoginForm"
]


class LoginForm(HTMLElement):
    text_icon: svg = Validator(logger=False, debug=True, required=True)
    text_label: CharLabel = Validator(logger=False, debug=True, required=True)
    text_input: TextInput = Validator(logger=False, debug=True, required=True)
    text_div: div = Validator(logger=False, debug=True, required=True)
    password_icon: svg = Validator(logger=False, debug=True, required=True)
    password_label: CharLabel = Validator(logger=False, debug=True, required=True)
    password_input: PasswordInput = Validator(logger=False, debug=True, required=True)
    password_div: div = Validator(logger=False, debug=True, required=True)
    button: SubmitButton = Validator(logger=False, debug=True, required=True)

    def __render__(
            self,
            *args,
            text_label,
            password_label,
            button_label,
            **kwargs
    ):
        self.text_icon = svg()
        self.text_label = CharLabel(label=text_label)
        self.text_input = TextInput(name="text", placeholder="Enter Username")
        self.text_div = self.html_tags.div(self.text_icon, self.text_label, self.text_input)
        self.password_icon = svg()
        self.password_label = CharLabel(label=password_label)
        self.password_input = PasswordInput(name="password", placeholder="Enter Password")
        self.password_div = self.html_tags.div(self.password_icon, self.password_label, self.password_input)
        self.button = SubmitButton(label=button_label, value="Submit")

        return form(self.text_div, self.password_div, self.button)


class RegisterForm(HTMLElement):
    text_icon: svg = Validator(logger=False, debug=True, required=True)
    text_label: CharLabel = Validator(logger=False, debug=True, required=True)
    text_input: TextInput = Validator(logger=False, debug=True, required=True)
    text_div: div = Validator(logger=False, debug=True, required=True)
    password_icon: svg = Validator(logger=False, debug=True, required=True)
    password_label: CharLabel = Validator(logger=False, debug=True, required=True)
    password_input: PasswordInput = Validator(logger=False, debug=True, required=True)
    password_div: div = Validator(logger=False, debug=True, required=True)
    confirm_password_icon: svg = Validator(logger=False, debug=True, required=True)
    confirm_password_label: CharLabel = Validator(logger=False, debug=True, required=True)
    confirm_password_input: PasswordInput = Validator(logger=False, debug=True, required=True)
    confirm_password_div: div = Validator(logger=False, debug=True, required=True)
    button: SubmitButton = Validator(logger=False, debug=True, required=True)

    def __render__(
            self,
            *args,
            text_label,
            password_label,
            button_label,
            **kwargs
    ):
        self.text_icon = svg()
        self.text_label = CharLabel(label=text_label)
        self.text_input = TextInput()
        self.text_div = div(self.text_icon, self.text_label, self.text_input)
        self.password_icon = svg()
        self.password_label = CharLabel(label=password_label)
        self.password_input = PasswordInput()
        self.password_div = div(self.password_icon, self.password_label, self.password_input)
        self.confirm_password_icon = svg()
        self.confirm_password_label = CharLabel(label=password_label)
        self.confirm_password_input = PasswordInput()
        self.confirm_password_div = div(
            self.confirm_password_icon,
            self.confirm_password_label,
            self.confirm_password_input
        )
        self.button = SubmitButton(label=button_label, value="Submit")

        return form(self.text_div, self.password_div, self.confirm_password_div, self.button)


class FieldSetLoginForm(HTMLElement):
    user_legend: CharLegend = Validator(logger=False, debug=True, required=True)
    user: TextInput = Validator(logger=False, debug=True, required=True)
    password: PasswordInput = Validator(logger=False, debug=True, required=True)
    button: SubmitButton = Validator(logger=False, debug=True, required=True)

    def __render__(
            self,
            *args,
            legend_label,
            text_label,
            password_label,
            button_label,
            **kwargs
    ):
        self.user_legend = CharLegend(legend_label)
        self.user = TextInput()
        self.password = PasswordInput()
        self.button = SubmitButton(label=button_label, value="Submit")
        return form(CharFieldSet(self.user_legend, self.user, self.password, self.button))


if __name__ == '__main__':
    seller_login = LoginForm(text_label="User ID", password_label="Password", button_label="Login")
    print(seller_login)
    # print(FieldSetLoginForm())



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

from dataclasses import dataclass

from uidom.dom.htmlelement import HTMLElement

__all__ = [
    "DateLabel",
    "DateInput",
]

@dataclass(eq=False)
class DateLabel(HTMLElement):
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def __render__(self, *args, **kwargs):
        return self.html_tags.label(*args, **kwargs)    


@dataclass(eq=False)
class DateInput(HTMLElement):
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __render__(self, *args, **kwargs):
        return self.html_tags.input_(*args, type="date", **kwargs)





# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


import typing as T
from functools import wraps

from starlette.background import BackgroundTask
from starlette.responses import Response

__all__ = [
    "DocumentResponse",
    "doc_response"
]

F = T.TypeVar("F", bound=T.Callable[..., T.Any])

class DocumentResponse(Response):
    media_type = "text/html"

    def __init__(
        self,
        html: T.Any,
        status_code: int = 200,
        headers: dict = None,
        media_type: str = None,
        background: BackgroundTask = None,
    ):
        self.html = html
        super().__init__(self.html, status_code, headers, media_type, background)

    def render(self, content: T.Any) -> bytes:
        if hasattr(content, "render"):
            content = content.render()
        content = super().render(content=content)
        return content


def doc_response(html_doc: F) -> T.Callable[..., DocumentResponse]:
    @wraps(html_doc)
    def decorator(*args, **kwargs) -> DocumentResponse:
        return DocumentResponse(html_doc(*args, **kwargs))
    return decorator



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.src import extension


class Python(extension.SingleTemplates):
    enable_right_delimiter_space = False
    enable_left_delimiter_space = True
    enable_space_in_between = False

    def render(self, indent="    ", pretty=True, xhtml=False):
        data = self._render([], 0, indent, pretty, xhtml)
        return "".join(data)


class Class(Python):
    self_dedent = True
    child_dedent = False
    left_delimiter = "class"
    right_delimiter = ":"

    def __init__(self, class_name, *dom_elements, bases=None):
        """
        :param class_name:
        :param dom_elements: [optional] template
        """

        super(Class, self).__init__(
            f"{class_name}",
            f"({', '.join(bases.split(' ')) if bases is not None else 'object'})",
            *dom_elements,
        )


class Args(Python):
    def __init__(self, func_name, *dom_elements):
        super(Args, self).__init__(f"{func_name}", "(self)", *dom_elements)


class Kwargs(Python):
    pass


class Def(Python):
    left_delimiter = "def"
    right_delimiter = ":"

    def __init__(self, func_name, *dom_elements, args=""):
        super(Def, self).__init__(f"{func_name}", f"({', '.join(['self', *args.split(' ')]) if any(args) else 'self'})",
                                  *dom_elements)
        if func_name == "__init__" and any(args):
            [self.add(SelfAttr(arg)) for arg in args.split(" ")]


class SelfAttr(Python):
    left_delimiter = ""
    right_delimiter = ""
    enable_left_delimiter_space = False

    def __init__(self, atr_name, *dom_elements):
        super(SelfAttr, self).__init__(f"self.{atr_name}", '', *dom_elements)


if __name__ == "__main__":
    import ast
    person = Class("Person",
                   # Def("__init__", args="name address"),
                   Def('__render__', args="*args **kwargs"),
                   bases="models.Model logging.Logger")
    print(person)



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom import htmlelement as htm
from uidom.dom.src.main import extension

__all__ = [
    "HtmlDocument"
]


class Meta(htm.HTMLElement):

    def __render__(self, **kwargs):
        return self.html_tags.meta(**kwargs)


class Head(htm.HTMLElement):

    def __render__(self, *args, **kwargs):
        return self.html_tags.head(*args, **kwargs)


class Body(htm.HTMLElement):

    def __render__(self, *args, **kwargs):
        return self.html_tags.body(*args, **kwargs)


class HtmlDocument(htm.HTMLElement):
    csrf_field = "X-CSRF-TOKEN"
    ensure_csrf_token_in_meta = True

    def __init__(self, *args, **kwargs):
        self.document = self
        super(HtmlDocument, self).__init__(*args, **kwargs)
        self._entry = self.body

    def __script__(self, element):
        ...

    def __checks__(self, element: extension.Tags) -> extension.Tags:
        if self.ensure_csrf_token_in_meta:
            token_element = element.get(name=self.csrf_field)
            if not token_element:
                raise AttributeError(f"{self.__class__.__qualname__} {self.csrf_field} must be set")
            if len(token_element) > 1:
                raise AssertionError(f"{self.__class__.__qualname__} {self.csrf_field} set at multiple places")
        return element

    def __render__(self, *args, head=None, body=None, common_head=None, common_body=None, **kwargs):
        common_head = [common_head] if not isinstance(common_head, list) else common_head
        common_body = [common_body] if not isinstance(common_body, list) else common_body
        head = [head] if not isinstance(head, list) else head
        body = [body] if not isinstance(body, list) else body

        # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        # ^Head Section
        # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        self.head = Head()

        if any(head):
            for _head_file in head:
                if isinstance(_head_file, extension.Tags):
                    self.head.add(_head_file)
                elif isinstance(_head_file, str):
                    self.head.add(self.html_tags.link(href=_head_file))
                elif isinstance(_head_file, dict):
                    self.head.add(self.html_tags.link(**_head_file))

        if any(common_head):
            _ = [self.head.add(_hd) for _hd in common_head if isinstance(_hd, extension.Tags)]
        # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        # $Head Section
        # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #

        # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        # ^Body Section
        # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        self.body = Body(*args, **kwargs)

        if any(body):
            for _body_file in body:
                if isinstance(_body_file, extension.Tags):
                    self.body.add(_body_file)
                elif isinstance(_body_file, str):
                    self.body.add(self.html_tags.script(src=_body_file))
                elif isinstance(_body_file, dict):
                    self.body.add(self.html_tags.script(**_body_file))

        if any(common_body):
            _ = [self.body.add(_bd) for _bd in common_body if isinstance(_bd, extension.Tags)]
        # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        # $Body Section
        # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ #
        if not self.head.get("meta", name="viewport"):
            viewport_meta = Meta(
                name="viewport",
                content="width=device-width, initial-scale=1, maximum-scale=1, user-scalable=no, minimal-ui"
            )
            self.head.add(viewport_meta)

        if not self.head.get("meta", charset="utf-8"):
            charset = Meta(charset="utf-8")
            self.head.add(charset)

        self.html = self.html_tags.html(self.head, self.body)
        doc = self.html_tags.DocType("html")
        return self.html_tags.ConcatTag(doc, self.html)




# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


import platform
import subprocess
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Union

import valio
from uidom.settings import WebAssets

__all__ = [
    "Command",
    "TailwindCommand"
]

def is_windows():
    if platform.system() == 'Windows':
        return True
    if platform.system().startswith('MINGW64_NT-'):
        return True
    return False


IS_WINDOWS = is_windows()

@dataclass
class Command(object):
    # command = valito.StringField(logger=False, debug=True)

    def run_command(self, *cmd, **kw) -> tuple[int, list[str]]:
        # self.command.logger.info(f'# {" ".join(cmd)}')

        if kw.get('shell'):
            while isinstance(cmd, list) or isinstance(cmd, tuple):
                cmd = cmd[0]
            # self.command.logger.debug(f'shell: {cmd}')
        with subprocess.Popen(cmd,
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE,
                              **kw) as p:
            if not p.stdout:
                raise Exception("fail to popen")
            while p.poll() is None:
                lines = []
                for line_bytes in iter(p.stdout.readline, b''):
                    line_bytes = line_bytes.rstrip()
                    try:
                        line = line_bytes.decode(encoding)
                    except (Exception,):
                        encoding = 'utf-8'
                        line = line_bytes.decode(encoding)
                    # self.command.logger.debug(line.rstrip())
                    lines.append(line)
                returncode = p.wait()
        if returncode != 0:
            raise subprocess.CalledProcessError(returncode, cmd)
        return p.returncode, lines


GROUPS = Union[str, list[str], None]


class GroupValidator(valio.Validator):
    annotation = GROUPS


class GroupField(valio.Field):
    validator = GroupValidator


# @dataclass
# class Group(Command):
#     groups = GroupField(logger=False, debug=True)
#
#     @groups.add_post_validator
#     def groups_as_list(self, group):
#         if isinstance(group, str):
#             return (group,)
#         return group
#
#     def exists(self) -> dict:
#         groups = dict()
#         for grp in self.group:
#             self.groups.logger.info(grp)
#             try:
#                 result = self.run_command('getent', 'group', grp)
#                 groups[grp] = result[1]
#                 self.groups.logger.info(result)
#             except (Exception,) as e:
#                 self.groups.logger.error(e)
#                 groups[grp] = False
#         return groups
#
#     def add(self, name):
#         ...
#
#     def remove(self, name):
#         ...
#
#     group: GROUPS = groups.validator


class TailwindValidator(valio.Validator):
    annotation = Union[str, list[str], None]


@dataclass
class TailwindCommand(Command):
    tailwindcss = TailwindValidator(debug=True, logger=False, default="tailwindcss")
    file_path: Union[str, Path]
    webassets: WebAssets
    input_file: Union[str, Path]
    output_file: Union[str, Path]
    minify: bool = False

    def __post_init__(self):
        self.root_dir = self.webassets.directory.ROOT_DIR
        self.project_dir = Path(self.file_path).parent
        self.input_file = self.root_dir / self.input_file
        self.output_file = self.webassets.static.css / self.output_file
        sys.exit(self.init_tailwind_project())

    def init_tailwind_project(self):
        if self.is_tailwindcss_available():
            tailwind_config_js = (self.root_dir / 'tailwind.config.js')
            if not tailwind_config_js.exists():
                print("initialising Tailwindcss Config")
                self.init_tailwind_config(init_dir=self.root_dir)
                if tailwind_config_js.exists():
                    with tailwind_config_js.open("w", encoding="utf-8") as tw:
                        tw.write(f'''
    module.exports = {{
        mode: "jit",
        content: {{
            files:[
            "../../{self.project_dir.relative_to(self.root_dir.parent.parent)}/*.{{html,py}}",
            "../../{self.project_dir.relative_to(self.root_dir.parent.parent)}/**/*.{{html,py}}",
            "../../{self.project_dir.relative_to(self.root_dir.parent.parent)}/**/**/*.{{html,py}}",
            ]
          }},
        plugins: [
            require('@tailwindcss/aspect-ratio'),
            require('@tailwindcss/forms'),
            require('@tailwindcss/line-clamp'),
            require('@tailwindcss/typography'),
            require('tailwindcss/colors'),
        ],
        theme: {{
            extend: {{}}
        }}
        
        }}''')
            if not self.input_file.exists() and tailwind_config_js.exists():
                with self.input_file.open("w", encoding="utf-8") as f:
                    f.write('''@tailwind base;\n@tailwind components;\n@tailwind utilities;''')

    def run_command(self, *cmd, **kw) -> tuple[int, list[str]]:
        # self.command.logger.info(f'# {" ".join(cmd)}')

        if kw.get('shell'):
            while isinstance(cmd, list) or isinstance(cmd, tuple):
                cmd = cmd[0]
            # self.command.logger.debug(f'shell: {cmd}')
        with subprocess.Popen(cmd,
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE,
                              **kw) as p:
            if not p.stdout:
                raise Exception("fail to popen")
            while p.poll() is None:
                lines = []
                for line_bytes in iter(p.stdout.readline, b''):
                    line_bytes = line_bytes.rstrip()
                    try:
                        line = line_bytes.decode(encoding)
                    except (Exception,):
                        encoding = 'utf-8'
                        line = line_bytes.decode(encoding)
                    # self.command.logger.debug(line.rstrip())
                    lines.append(line)
                returncode = p.wait()
        if returncode != 0:
            raise subprocess.CalledProcessError(returncode, cmd)
        return p.returncode, lines

    def is_tailwindcss_available(self):
        output = subprocess.run([self.tailwindcss], cwd=self.root_dir.as_posix().encode() if not IS_WINDOWS else str(self.root_dir).encode())
        print("tailwindcss command :> ", output)
        return output

    def init_tailwind_config(self, init_dir: Path):
        print("trying to init tailwindcss config")
        output = subprocess.run([self.tailwindcss, "init"], cwd=init_dir.as_posix().encode() if not IS_WINDOWS else str(self.root_dir).encode())
        print("intialisation > ", bool(output))

    def start(self):
        print("cwd", self.root_dir)
        print("input_file > ", self.input_file)
        print("output_file > ", self.output_file)
        try:
            output = subprocess.run([
                self.tailwindcss,
                '-i', self.input_file.relative_to(self.root_dir),
                "-o", self.output_file.relative_to(self.root_dir),
                f"--{(self.minify and 'minify') or 'watch'}",
                ],
                cwd=self.root_dir
            )
            return output
        except (Exception,) as e:
            print(e)

#
# if __name__ == "__main__":
#     print(Group(["asdf", "kubctrl"]).exists())



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.elements import Button
from uidom.dom.htmlelement import HTMLElement
from uidom.dom.icons import Icons
from uidom.dom.links import HtmxLink
from uidom.dom.tailwindcss import Absolute, InlineBlock, InlineFlex, Relative


class DropdownItem(HTMLElement):

    def __render__(self, *args, item_link, target_id, item_name=None, **kwargs):
        self.htmx_link = HtmxLink(*args,
                                  item_link=item_link,
                                  target_id=target_id,
                                  item_name=item_name,
                                  **kwargs)
        self.icon = Icons()
        drop_link = self.html_tags.div(self.icon, self.htmx_link)
        drop_link["class"] = ""
        return drop_link


class DropdownElement(HTMLElement):

    def __render__(self, *args, value="Options", **kwargs):
        self.icon = Icons()
        button = Button("Options", self.icon, value=value)
        button["class"] = ""
        button = InlineFlex(button)
        self.button = button
        self.absolute = self.html_tags.div(*args, **kwargs)
        self.absolute["class"] = ""
        self.absolute = Absolute(self.absolute)
        dropper = self.html_tags.div(self.html_tags.div(self.button), self.absolute)
        dropper["class"] = ""
        return InlineBlock(Relative(dropper))


if __name__ == '__main__':
    account = DropdownItem(item_link="/account", target_id="#account-data", item_name="account")
    account["class"] += " py-1"
    dropdown = DropdownElement(account)
    print(dropdown)



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.htmlelement import HTMLElement


class Top(HTMLElement):

    def __render__(self, elem, sign="", position="0"):
        elem["class"] += f"{sign}top-{position}"
        return elem


class Right(HTMLElement):

    def __render__(self, elem, sign="", position="0"):
        elem["class"] += f"{sign}right-{position}"
        return elem


class Bottom(HTMLElement):

    def __render__(self, elem, sign="", position="0"):
        elem["class"] += f"{sign}bottom-{position}"
        return elem


class Left(HTMLElement):

    def __render__(self, elem, sign="", position="0"):
        elem["class"] += f"{sign}left-{position}"
        return elem


class Inset(HTMLElement):

    def __render__(self, elem, sign="", position="0"):
        elem["class"] += f"{sign}inset-{position}"
        return elem



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from functools import wraps

from .flexbox import Flex, FlexColumn, FlexRow
from .gridbox import (ColumnEnd, Columns, ColumnSpan, ColumnStart, Grid,
                      GridFlow, RowEnd, Rows, RowSpan, RowStart)
from .layout import Large, Medium, Small
from .responsive import Focus, Hover


def flex(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        return Flex(func(*args, **kwargs))
    return wrapper


def flex_col(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        return FlexColumn(func(*args, **kwargs))
    return wrapper


def flex_row(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        return FlexRow(func(*args, **kwargs))
    return wrapper


def grid(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        return Grid(func(*args, **kwargs))
    return wrapper


def grid_flow(flow="row"):
    def fn(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return GridFlow(func(*args, **kwargs), flow=flow)
        return wrapper
    return fn


def hover(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        return Hover(func(*args, **kwargs))
    return wrapper


def focus(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        return Focus(func(*args, **kwargs))
    return wrapper


def sm(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        return Small(func(*args, **kwargs))
    return wrapper


def md(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        return Medium(func(*args, **kwargs))
    return wrapper


def lg(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        return Large(func(*args, **kwargs))
    return wrapper


def column(columns="12"):
    def fn(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return Columns(func(*args, **kwargs), columns=columns)
        return wrapper
    return fn


def column_start(start="1"):
    def fn(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return ColumnStart(func(*args, **kwargs), start=start)
        return wrapper
    return fn


def column_span(span="1"):
    def fn(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return ColumnSpan(func(*args, **kwargs), span=span)
        return wrapper
    return fn


def column_end(end="1"):
    def fn(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return ColumnEnd(func(*args, **kwargs), end=end)
        return wrapper
    return fn


def row(rows="6"):
    def fn(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return Rows(func(*args, **kwargs), rows=rows)
        return wrapper
    return fn


def row_start(start="1"):
    def fn(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return RowStart(func(*args, **kwargs), start=start)
        return wrapper
    return fn


def row_span(span="1"):
    def fn(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return RowSpan(func(*args, **kwargs), span=span)
        return wrapper
    return fn


def row_end(end="1"):
    def fn(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return RowEnd(func(*args, **kwargs), end=end)
        return wrapper
    return fn



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


__license__ = '''
This file is part of Dominate.

Dominate is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

Dominate is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with Dominate.  If not, see
<http://www.gnu.org/licenses/>.
'''

try:
    basestring = basestring
except NameError:  # py3
    basestring = str
    unicode = str


class dom1core(object):
    '''
    Implements the Document Object Model (Core) Level 1

    http://www.w3.org/TR/1998/REC-DOM-Level-1-19981001/
    http://www.w3.org/TR/1998/REC-DOM-Level-1-19981001/level-one-core.html
    '''

    @property
    def parentNode(self):
        '''
        DOM API: Returns the parent tag of the current element.
        '''
        return self.parent

    def getElementById(self, id):
        '''
        DOM API: Returns single element with matching id value.
        '''
        results = self.get(id=id)
        if results is None:
            raise ValueError('No tags with id "%s".' % id)
        elif len(results) > 1:
            raise ValueError('Multiple tags with id "%s".' % id)
        elif results:
            return results[0]
        else:
            return None

    def getElementsByTagName(self, name):
        '''
        DOM API: Returns all tags that match name.
        '''
        if isinstance(name, basestring):
            return self.get(name.lower())
        else:
            return None

    def appendChild(self, obj):
        '''
        DOM API: Add an item to the end of the children list.
        '''
        self.add(obj)
        return self



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.htmlelement import HTMLElement


class Z0(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "z-0"
        return elem


class Z10(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "z-10"
        return elem


class Z20(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "z-20"
        return elem


class Z30(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "z-30"
        return elem


class Z40(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "z-40"
        return elem


class Z50(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "z-50"
        return elem


class ZAuto(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "z-auto"
        return elem



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


import typing
from dataclasses import dataclass


class BaseState(dict):
    """
    Base States
    """


@dataclass
class BaseDBManager(object):
    database: typing.Any
    password: typing.Any
    table: typing.Any

    async def post(self, *args, **kwargs: BaseState):
        ...

    async def put(self, *args, **kwargs: BaseState):
        ...

    async def delete(self, *args, **kwargs: BaseState):
        ...

    async def get(self, *args, **kwargs: BaseState):
        ...


@dataclass
class BaseStateManager(object):
    state_names = ["id", "name"]  # just a placeholder
    state_db: BaseDBManager

    async def client_state_handler(self, **client_states: BaseState):
        if not all(state in client_states for state in self.state_names):
            return

        if not self.state_db.get(self.state_names, **client_states):
            return

        # updating client_states' value for state_names
        client_states = await self.update_states(self.state_names, **client_states)

        # updating state_db's value for state_names
        return await self.state_db.put(self.state_names, **client_states)

    async def server_state_handler(self, **server_states: BaseState):
        if not all(state in server_states for state in self.state_names):
            server_states = await self.create_states(self.state_names, **server_states)

        if self.state_db.get(self.state_names, **server_states):
            return server_states

        return await self.state_db.post(self.state_names, **server_states)

    async def create_states(self, state_names, **states):  # noqa
        return states

    async def update_states(self, state_names,  **states):  # noqa
        return states


@dataclass
class BaseEventManager(object):
    event_name = "events"
    states_manager: BaseStateManager
    events_db: BaseDBManager

    async def dispatch_event(self, **states):  # noqa
        return states

    # async def client_event_manager(self, **client_states: BaseState):
    #     client_states = await self.states_manager.client_state_handler(**client_states)
    #     if not client_states or self.event_name not in client_states:
    #         return client_states
    #     client_states = await self.events_store.get(self.event_name, **client_states)
    #     async for client_event in self.events_store.get(self.event_name, **client_states):
    #         await self.dispatch_event(**client_event)

    async def server_event_manager(self, **current_data: BaseState):
        ...



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


import typing as T
from dataclasses import dataclass, field

from uidom.dom.htmlelement import HTMLElement
from valio import StringValidator

__all__ = [
    # integer html
    "FloatLabel",
    "FloatLegend",
    "FloatInput",
    "FloatNumberInput",
    "FloatRangeInput",
]


@dataclass(eq=False)
class FloatLabel(HTMLElement):
    label: str = StringValidator(logger=False, debug=True)

    def __post_init__(self, *args, **kwargs):
        super(FloatLabel, self).__init__(self.label, *args, **kwargs)

    def __render__(self, label, *args, **kwargs):
        return self.html_tags.label(label, *args, **kwargs)


@dataclass(eq=False)
class FloatLegend(HTMLElement):
    label: str = StringValidator(logger=False, debug=True)

    def __post_init__(self, *args, **kwargs):
        super(FloatLegend, self)._init__(self.label, *args, **kwargs)

    def __render__(self, label, *args, **kwargs):
        return self.html_tags.legend(label, *args, **kwargs)


min_field = StringValidator(logger=False, debug=True)
max_field = StringValidator(logger=False, debug=True)


@dataclass(eq=False)
class FloatInput(HTMLElement):
    
    @max_field.add_validator
    @min_field.add_validator
    def cast2float(self, value):
        if value not in ["null", None]:
            try:
                float(value)
            except TypeError as e:
                raise ValueError(f"{value} is not a float value") from e
    

    name: str = StringValidator(logger=False, debug=True)
    type: str = StringValidator(in_choice=["number", "range"], logger=False, debug=True, default="number")
    placeholder: str = StringValidator(logger=False, debug=True)
    min: str = min_field
    max: str = max_field
    pattern: str = StringValidator(logger=False, debug=True)

    def __post_init__(self, *args, **kwargs):
        super(FloatInput, self).__init__(*args,
                                         name=self.name,
                                         placeholder=self.placeholder,
                                         type=self.type,
                                         min=self.min,
                                         max=self.max,
                                         pattern=self.pattern,
                                         **kwargs
                                         )

    def __render__(
            self,
            *args,
            name,
            placeholder="",
            type,
            min,
            max,
            pattern,
            **kwargs
    ):
        return self.html_tags.input_(
            *args,
            type=type,
            name=name,
            placeholder=placeholder,
            min=min if min not in ["null", None] else False,
            max=max if max not in ["null", None] else False,
            pattern=pattern if pattern not in ["null", None] else False,
            **kwargs,
        )


@dataclass(eq=False)
class FloatNumberInput(HTMLElement):

    name: str
    placeholder: str
    min: T.Union[str, None] = None
    max: T.Union[str, None] = None
    pattern: T.Union[str, None] = None

    def __post_init__(self, *args, **kwargs):
        super(FloatNumberInput, self).__init__(*args,
                                               name=self.name,
                                               placeholder=self.placeholder,
                                               min=self.min,
                                               max=self.max,
                                               pattern=self.pattern,
                                               **kwargs
                                               )

    def __render__(
            self,
            *args,
            name,
            placeholder="",
            min,
            max,
            pattern,
            **kwargs
    ):
        return FloatInput(
            name=name,
            placeholder=placeholder,
            type="number",
            min=min,
            max=max,
            pattern=pattern,
        )


@dataclass(eq=False)
class FloatRangeInput(HTMLElement):
    name: str
    placeholder: str
    min: T.Union[str, None] = None
    max: T.Union[str, None] = None
    pattern: T.Union[str, None] = None

    def __post_init__(self, *args, **kwargs):
        super(FloatRangeInput, self).__init__(*args,
                                              name=self.name,
                                              placeholder=self.placeholder,
                                              min=self.min,
                                              max=self.max,
                                              pattern=self.pattern,
                                              **kwargs
                                              )

    def __render__(
            self,
            *args,
            name,
            placeholder="",
            min,
            max,
            pattern,
            **kwargs
    ):
        return FloatInput(
            name=name,
            placeholder=placeholder,
            type="range",
            min=min,
            max=max,
            pattern=pattern,
        )


if __name__ == '__main__':
    from valio import Pattern, SetOf
    print(FloatRangeInput(name="test", placeholder="xyz"))
    print(FloatInput(name="test", type="number", placeholder="space",
                     pattern=SetOf(Pattern("0-9", count_min=1, count_max=6)).pattern, min="100"))



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from .css import *
from .flexbox import *
from .gridbox import *
from .layout import *
from .placements import *
from .responsive import *



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from .widgets import *  # isort: skip
from .htmlelement import *  # isort: skip
from .htmldocument import *  # isort: skip
from .xcomponent_js import *  # isort: skip
# from .xshadowcomponent import *  # isort: skip
from .css import *  # isort: skip
from .elements import *  # isort: skip 
from .htmlelement import *  # isort: skip
from .icons import *  # isort: skip
from .jinja import *  # isort: skip
from .links import *  # isort: skip
from .src import *  # isort: skip
# from .tailwindcss import *  # isort: skip
from .vue import *  # isort: skip
from .html_to_dom import *  # isort: skip

# https://github.com/microsoft/vscode/issues/35350#issuecomment-1093627529 
# here is the need to include isort skip.



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.htmlelement import HTMLElement


class OverScrollAuto(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "overscroll-auto"
        return elem


class OverScrollContain(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "overscroll-contain"
        return elem


class OverScrollNone(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "overscroll-none"
        return elem


class OverScrollYAuto(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "overscroll-y-auto"
        return elem


class OverScrollYContain(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "overscroll-y-contain"
        return elem


class OverScrollYNone(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "overscroll-y-none"
        return elem


class OverScrollXAuto(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "overscroll-x-auto"
        return elem


class OverScrollXContain(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "overscroll-x-contain"
        return elem


class OverScrollXNone(HTMLElement):

    def __render__(self, elem):
        elem["class"] += "overscroll-x-none"
        return elem



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT



import os
from collections.abc import Iterable
from typing import List, Union

__all__ = [
    "make_paths"
]


def path_exists(path):
    if not os.path.isabs(path):
        path = os.path.abspath(path)
    path_exist = os.path.exists(path)
    return path_exist


class SubdirOrFile(object):
    def __init__(self, sub_directory_or_file: Union[str, List[str]] = None):
        self._sub_dir_or_file = (
            (
                map(lambda x: os.path.abspath(x), sub_directory_or_file)
                if isinstance(sub_directory_or_file, Iterable)
                else [os.path.abspath(sub_directory_or_file)]
            )
            if sub_directory_or_file is not None
            else [sub_directory_or_file]
        )

        self._iter_sub_dir_or_file = None

    def sub_dir_or_file(self):
        return self._sub_dir_or_file

    def __iter__(self):
        for sub_dir_or_file in self.sub_dir_or_file():
            yield sub_dir_or_file

    def __next__(self):
        if self._iter_sub_dir_or_file is None:
            self._iter_sub_dir_or_file = iter(self)
        return next(self._iter_sub_dir_or_file)

    def _get_path(self, sub_directory_or_file=None, directory=None):
        sub_directory_or_file = (
            sub_directory_or_file
            if isinstance(sub_directory_or_file, Iterable)
            else [sub_directory_or_file]
            if sub_directory_or_file is not None
            else iter(self)
        )

        if directory is None:
            paths = sub_directory_or_file
        else:
            paths = list()
            for sub_dir in sub_directory_or_file:
                if sub_dir:
                    if isinstance(directory, Iterable):
                        for direct in directory:
                            paths.extend(os.path.join(direct, sub_dir))
                    else:
                        paths = os.path.join(directory, sub_dir)
        return paths

    @staticmethod
    def existing(paths):
        if any(paths):
            for p in paths:
                if path_exists(p):
                    yield p
        else:
            yield paths


class MakePath(SubdirOrFile):
    def __init__(self, sub_directory_or_file, directory=None, create_file=False):
        super(MakePath, self).__init__(sub_directory_or_file=sub_directory_or_file)
        self._directory = directory
        self._create_file = create_file

        self._all_paths = self._get_path(sub_directory_or_file, self._directory)
        self._paths_that_exits = set(self.existing(self._all_paths))
        self._paths_to_create = set(self._all_paths).difference(self._paths_that_exits)
        self._paths_created = set()

    def paths_exists(self, path_name=None):
        if path_name:
            return path_name in self._paths_that_exits
        return self._paths_that_exits

    def paths_to_create(self):
        return self._paths_to_create

    def paths_created(self):
        return self._paths_created

    def make_path(self):
        paths_created = list()

        for path in self._paths_to_create:
            if "." not in os.path.basename(path):
                if not path_exists(path):
                    if path == "__main__":
                        continue
                    os.makedirs(path)
                    paths_created.append(path)
            elif self._create_file:
                path_dir = os.path.dirname(path)
                if not path_exists(path_dir):
                    os.makedirs(path_dir)
                    paths_created.append(path_dir)

                with open(path, mode="a"):
                    paths_created.append(path)

        for path in paths_created:
            self._paths_to_create.remove(path)
            self._paths_created.add(path)
            if path not in self._paths_that_exits:
                self._paths_that_exits.add(path)

    def __str__(self):
        return "<MakePath({files})>".format(files=self._all_paths)


def make_paths(dirs_list):
    paths = MakePath(dirs_list)
    paths.make_path()
    return paths.paths_created()



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.src import extension


class Sphinx(extension.SingleTemplates):
    self_dedent = True
    left_delimiter = ''
    right_delimiter = ''
    enable_left_delimiter_space = False
    enable_right_delimiter_space = False


class Heading(Sphinx):
    child_dedent = True

    def __init__(self, heading_name, *dom_elements):
        """
        :param heading_name:
        :param dom_elements:
        """
        super(Heading, self).__init__("", heading_name, Sphinx("", "=" * len(heading_name)), *dom_elements)


class VarName(Sphinx):
    left_delimiter = "`"
    right_delimiter = "`"

    def __init__(self, type_text, *dom_elements):
        """
        :param type_text:
        :param dom_elements: [optional] template
        """
        super(VarName, self).__init__("", type_text, *dom_elements)


class Class(Sphinx):
    left_delimiter = ":"
    right_delimiter = ":"
    self_dedent = True

    def __init__(self, template_text, *dom_elements):
        """
        :param template_text:
        :param dom_elements: [optional] template
        """
        super(Class, self).__init__("class", template_text, *dom_elements)


class Function(Sphinx):
    left_delimiter = ""

    def __init__(self, template_text, *dom_elements):
        """
        :param template_text:
        :param dom_elements: [optional] template
        """
        super(Function, self).__init__(":class:", template_text, *dom_elements)


if __name__ == "__main__":
    var_name = VarName("something")
    class_name = Class("BaseClass", var_name)
    print(Heading("Hello0ijj", class_name, Function("ls")))



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


'''
Utility classes for creating dynamic html documents
'''

__license__ = '''
This file is part of Dominate.

Dominate is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

Dominate is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with Dominate.  If not, see
<http://www.gnu.org/licenses/>.
'''

import re

from uidom.dom.src.dom_tag import dom_tag

try:
    basestring = basestring
except NameError:
    basestring = str
    unichr = chr


def include(f):
    '''
    includes the contents of a file on disk.
    takes a filename
    '''
    fl = open(f, 'r')
    data = fl.read()
    fl.close()
    return raw(data)


def system(cmd, data=None):
    '''
    pipes the output of a program
    '''
    import subprocess
    s = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stdin=subprocess.PIPE)
    out, err = s.communicate(data)
    return out.decode('utf8')


def escape(data, quote=True):  # stolen from std lib cgi
    '''
    Escapes special characters into their html entities
    Replace special characters "&", "<" and ">" to HTML-safe sequences.
    If the optional flag quote is true, the quotation mark character (")
    is also translated.

    This is used to escape content that appears in the body of an HTML document
    '''
    data = data.replace("&", "&amp;")  # Must be done first!
    data = data.replace("<", "&lt;")
    data = data.replace(">", "&gt;")
    if quote:
        data = data.replace('"', "&quot;")
    return data


_unescape = {
    'quot': 34,
    'amp': 38,
    'lt': 60,
    'gt': 62,
    'nbsp': 32,
    # more here
    # http://www.w3.org/TR/html4/sgml/entities.html
    'yuml': 255,
}
str_escape = escape


def unescape(data):
    '''
    unescapes html entities. the opposite of escape.
    '''
    cc = re.compile(r'&(?:(?:#(\d+))|([^;]+));')

    result = []
    m = cc.search(data)
    while m:
        result.append(data[0:m.start()])
        d = m.group(1)
        if d:
            d = int(d)
            result.append(unichr(d))
        else:
            d = _unescape.get(m.group(2), ord('?'))
            result.append(unichr(d))

        data = data[m.end():]
        m = cc.search(data)

    result.append(data)
    return ''.join(result)


_reserved = ";/?:@&=+$, "
_replace_map = dict((c, '%%%2X' % ord(c)) for c in _reserved)


def url_escape(data):
    return ''.join(_replace_map.get(c, c) for c in data)


def url_unescape(data):
    return re.sub('%([0-9a-fA-F]{2})',
                  lambda m: unichr(int(m.group(1), 16)), data)


class lazy(dom_tag):
    '''
    delays function execution until rendered
    '''

    def __new__(_cls, *args, **kwargs):
        '''
        Need to reset this special method or else
        dom_tag will think it's being used as a dectorator.

        This means lazy() can't be used as a dectorator, but
        thinking about when you might want that just confuses me.
        '''
        return object.__new__(_cls)

    def __init__(self, func, *args, **kwargs):
        super(lazy, self).__init__()
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def _render(self, sb, *a, **kw):
        r = self.func(*self.args, **self.kwargs)
        sb.append(str(r))


# TODO rename this to raw?
class dom_text(dom_tag):
    '''
    Just a string. useful for inside context managers
    '''
    is_pretty = False
    is_inline = False

    def __init__(self, _text, escape=True):
        super(dom_text, self).__init__()
        if escape:
            self.text = str_escape(_text)
        else:
            self.text = _text

    def _render(self, sb, *a, **kw):
        sb.append(self.text)
        return sb


def raw(s):
    '''
    Inserts a raw string into the DOM. Unsafe.
    '''
    return dom_text(s, escape=False)



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom import *

__all__ = [
    "XLinks",
    "x_links"
]


class XLinks(CustomElement):

    def __render__(self, tag_name):
        return template(div(x_component=tag_name))


x_links = XLinks(tag_name="links")






# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.htmlelement import HTMLElement
from valio import Validator


class Gap(HTMLElement):

    def __render__(self, elem, gap=""):
        elem["class"] += f" gap-{gap}"
        return elem


class GapX(HTMLElement):

    def __render__(self, elem, x="px"):
        elem["class"] += f" gap-x-{x}"
        return elem


class GapY(HTMLElement):

    def __render__(self, elem, y="px"):
        elem["class"] += f" gap-y-{y}"
        return elem


class JustifyContent(HTMLElement):
    justify: str = Validator(logger=False,
                             debug=True,
                             in_choice=["start", "end", "center", "between", "around", "evenly"])

    def __render__(self, elem, justify="center"):
        self.justify = justify
        elem["class"] += f" justify-{justify}"
        return elem


class JustifyItem(HTMLElement):
    justify: str = Validator(logger=False,
                             debug=True,
                             in_choice=["start", "end", "center", "stretch"])

    def __render__(self, elem, justify="center"):
        self.justify = justify
        elem["class"] += f" justify-items-{justify}"
        return elem


class JustifySelf(HTMLElement):
    justify: str = Validator(logger=False,
                             debug=True,
                             in_choice=["start", "end", "center", "stretch", "auto"])

    def __render__(self, elem, justify="center"):
        self.justify = justify
        elem["class"] += f" justify-self-{justify}"
        return elem


class AlignContent(HTMLElement):
    align: str = Validator(logger=False,
                           debug=True,
                           in_choice=["start", "end", "center", "between", "around", "evenly"])

    def __render__(self, elem, align="center"):
        self.align = align
        elem["class"] += f" content-{align}"
        return elem


class AlignItem(HTMLElement):
    align: str = Validator(logger=False,
                           debug=True,
                           in_choice=["start", "end", "center", "stretch", "baseline"])

    def __render__(self, elem, align="center"):
        self.align = align
        elem["class"] += f" items-{align}"
        return elem


class AlignSelf(HTMLElement):
    align: str = Validator(logger=False,
                           debug=True,
                           in_choice=["start", "end", "center", "stretch", "auto"])

    def __render__(self, elem, align="center"):
        self.align = align
        elem["class"] += f" self-{align}"
        return elem


class PlaceContent(HTMLElement):
    place: str = Validator(logger=False,
                           debug=True,
                           in_choice=["start", "end", "center", "between", "around", "evenly", "stretch"])

    def __render__(self, elem, place="center"):
        self.place = place
        elem["class"] += f" place-content-{place}"
        return elem


class PlaceItem(HTMLElement):
    place: str = Validator(logger=False,
                           debug=True,
                           in_choice=["start", "end", "center", "stretch", "baseline"])

    def __render__(self, elem, place="center"):
        self.place = place
        elem["class"] += f" place-items-{place}"
        return elem


class PlaceSelf(HTMLElement):
    place: str = Validator(logger=False,
                           debug=True,
                           in_choice=["start", "end", "center", "stretch", "auto"])

    def __render__(self, elem, place="center"):
        self.place = place
        elem["class"] += f" place-self-{place}"
        return elem



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


import os
import typing

from jinja2.utils import htmlsafe_json_dumps
from uidom.dom.src.dom1core import dom1core
from uidom.dom.src.dom_tag import dom_tag, unicode
from uidom.dom.src.utils.dom_util import escape

__all__ = ["SingleTemplates", "DoubleTemplates", "DoubleTags", "SingleTags", "StyleTags"]


class Tags(dom_tag, dom1core):
    left_delimiter = "<"
    right_delimiter = ">"
    self_dedent = False
    child_dedent = False
    render_tag = True
    new_line_at_end = False
    SELF_DEDENT = "self_dedent"
    CHILD_DEDENT = "child_dedent"
    OPEN_TAG = "open_tag"
    CLOSE_TAG = "close_tag"
    RENDER_TAG = "render_tag"
    NEW_LINE_AT_CHILD_END = "new_line_at_child_end"
    file_extension = ".html"

    def __init__(self, *args, **kwargs):
        if any(args):
            msg = f"can only pass {dom_tag!r} or {str!r} types in arguments, got {args!r} instead"
            if not all(map(lambda x: isinstance(x, (dom_tag, str)), args)):
                raise TypeError(msg)
        super(Tags, self).__init__(*args, **kwargs)

    def _render_open_tag(self, sb, indent_level, indent_str, pretty, xhtml, name, open_tag):
        if open_tag:
            sb.append("%s" % open_tag)
        else:
            # open tag is absent
            sb.append(self.left_delimiter)
            sb.append(name)
            sb = self._render_attribute(sb, indent_level, indent_str, pretty, xhtml)
            sb.append(''.join(["/", self.right_delimiter]) if self.is_single and xhtml else self.right_delimiter)
        return sb

    def _render_attribute(self, sb, indent_level, indent_str, pretty, xhtml):
        for attribute, value in sorted(self.attributes.items()):
            if value is not False and value not in [None]:  # False values must be omitted completely
                if not isinstance(value, (typing.MutableMapping, typing.MutableSequence)):
                    sb.append(' %s="%s"' % (attribute, escape(unicode(value), True)))
                else:
                    value = htmlsafe_json_dumps(value)
                    sb.append(" %s='%s'" % (attribute, escape(unicode(value), True)))
            if value in [None]:  # minified xhtml attributes are added
                sb.append(' %s' % attribute)
        return sb

    def _render_close_tag(self, sb, name, close_tag):
        if close_tag:
            sb.append("%s" % close_tag)
        else:
            sb.append(''.join([self.left_delimiter, "/"]))
            sb.append(name)
            sb.append(self.right_delimiter)
        return sb

    @staticmethod
    def _new_line_and_inline_handler(sb, indent_level, indent_str, pretty, is_inline):
        if pretty and not is_inline:
            is_inline = False
            sb.append("\n")
            sb.append(indent_str * indent_level)
        return sb, is_inline

    @staticmethod
    def _dedent_handler(dedent, indent_level):
        if dedent:
            indent_level -= 1
        return indent_level

    @staticmethod
    def clean_attribute(attribute):
        """
        Normalize attribute names for shorthand and work arounds for limitations
        in Python's syntax. Extended it to support VueJS, HTMX and AngularJS.
        """

        # Shorthand
        attribute = {
            'cls': 'class',
            'className': 'class',
            'class_name': 'class',
            'fr': 'for',
            'html_for': 'for',
            'htmlFor': 'for',
        }.get(attribute, attribute)

        # Workaround for Python's reserved words
        if len(attribute) >= 2:
            if attribute[0] == '_' and attribute[1] != "_":
                attribute = attribute[1:]

        # Workaround for dash plus support for VueJS, HTMX, Unpoly and AngularJS
        special_prefix = any(
            [attribute.startswith(x) for x in ('data_', 'aria_', 'x_', 'v_', 'ng_', "hx_", "__", "ws__", "up_",
                                               "remove_me")])
        if attribute in {'http_equiv'} or special_prefix:
            attribute = attribute.replace('_', '-')
            attribute = attribute.replace("--", ":")
            attribute = attribute.replace("v-bind-", ":")
            attribute = attribute.replace("v-bind", "")
            attribute = attribute.replace("x-bind-", ":")
            attribute = attribute.replace("x-bind", "")
            attribute = attribute.replace("x-transition-", "x-transition:")
            attribute = attribute.replace("v-on:", "@")
            attribute = attribute.replace("v-on-", "@")
            attribute = attribute.replace("x-on:", "@")
            attribute = attribute.replace("x-on-", "@")
            attribute = attribute.replace("-dot-", ".")

        # Workaround for colon
        if attribute.split('_')[0] in ('xlink', 'xml', 'xmlns'):
            attribute = attribute.replace('_', ':', 1)

        return attribute

    def _clean_name(self, name):
        # Workaround for python keywords and standard classes/methods
        # (del, object, input)
        if any(name):  # to handle the case when tagname = ""
            if name[-1] == '_':
                name = name[:-1]
        return name

    def _render_children(self, sb, indent_level, indent_str, pretty, xhtml):
        inline = True
        orig_indent = indent_level
        self_render_tag = self.attributes.pop(Tags.RENDER_TAG,
                                              getattr(self, Tags.RENDER_TAG)
                                              if hasattr(self, Tags.RENDER_TAG) else True)
        for child in self.children:
            if isinstance(child, dom_tag):

                child_self_dedent = child.attributes.pop(Tags.SELF_DEDENT,
                                                         getattr(child, Tags.SELF_DEDENT)
                                                         if hasattr(child, Tags.SELF_DEDENT) else False)

                # checks if the self.child_dedent or self.self_dedent if true and dedent the tags
                if pretty and not child.is_inline:
                    inline = False
                    dedent = child_self_dedent
                    if dedent:
                        if indent_level > orig_indent - 1:
                            indent_level = self._dedent_handler(dedent, indent_level)

                if self_render_tag:
                    sb, inline = self._new_line_and_inline_handler(
                        sb, indent_level, indent_str, pretty, inline and self.is_inline)

                child._render(sb, indent_level, indent_str, pretty, xhtml)

            else:
                if any(child):
                    if not (pretty and not self.is_inline):
                        sb.append(unicode(child))
                    else:
                        inline = False
                        sb, inline = self._new_line_and_inline_handler(
                            sb, indent_level, indent_str, pretty, inline)
                        sb.append(unicode(child))

            # new_line_at_end caters to DOCTYPE Tag as its wrapped in an empty wrapper
            new_line_at_child_end = self.attributes.pop(Tags.NEW_LINE_AT_CHILD_END,
                                                        getattr(self, Tags.NEW_LINE_AT_CHILD_END)
                                                        if hasattr(self, Tags.NEW_LINE_AT_CHILD_END) else False)

            # new_line_at_child_end caters to ConcatTag as its an empty wrapper and children needs
            # '\n' new line support except the last one
            if new_line_at_child_end and self.children[-1] != child:
                sb, _ = self._new_line_and_inline_handler(sb, indent_level, indent_str, pretty,
                                                          inline and self.is_inline)
        return inline

    def _render(self, sb, indent_level=1, indent_str="  ", pretty=True, xhtml=False):
        open_tag = self.attributes.pop(Tags.OPEN_TAG, False)
        close_tag = self.attributes.pop(Tags.CLOSE_TAG, False)
        pretty = pretty and self.is_pretty and not self.is_inline
        # name = self._clean_name(getattr(self, 'tagname', type(self).__name__))
        self_child_dedent = self.attributes.pop(Tags.CHILD_DEDENT,
                                                getattr(self, Tags.CHILD_DEDENT)
                                                if hasattr(self, Tags.CHILD_DEDENT) else False)

        self_render_tag = self.attributes.pop(Tags.RENDER_TAG,
                                              getattr(self, Tags.RENDER_TAG)
                                              if hasattr(self, Tags.RENDER_TAG) else True)

        dedent = (not self_render_tag) or self_child_dedent
        if pretty and dedent:
            indent_level = self._dedent_handler(dedent, indent_level)

        if self_render_tag:
            name = self._clean_name(getattr(self, 'tagname', type(self).__name__))
            self._render_open_tag(sb, indent_level, indent_str, pretty, xhtml, name, open_tag)

        inline = self._render_children(sb,
                                       indent_level + 1 if not self.is_single or not self_child_dedent else indent_level,
                                       indent_str, pretty, xhtml)
        inline = self.is_inline and inline
        if self_render_tag:
            if not self.is_single:
                sb, inline = self._new_line_and_inline_handler(sb, indent_level, indent_str, pretty, inline)
                name = self._clean_name(getattr(self, 'tagname', type(self).__name__))
                self._render_close_tag(sb, name, close_tag)

        return sb

    def save(
            self,
            file_name: str = None,
            template_folder: str = "static",
            html_template=None,
            current_dir=False,
    ):
        html_template = html_template or self.render()
        current_folder = os.getcwd()
        doc_folder = current_folder if current_dir else os.path.dirname(current_folder)
        template_folder = os.path.join(doc_folder, template_folder)
        file_name = file_name or str(self.__class__.__name__)
        if self.file_extension is None:
            raise ValueError(f"can not save file with {self.file_extension!r} type extension")
        template_file = (
            file_name
            if file_name.endswith(self.file_extension)
            else ".".join([file_name, self.file_extension[1:]])
        )
        if not os.path.exists(template_folder):
            os.makedirs(template_folder)

        template_file_path = os.path.join(template_folder, template_file)

        if not os.path.exists(template_file_path):
            with open(template_file_path, "w+") as f:
                f.write(html_template)
        else:
            with open(template_file_path, "r") as temp:
                old_html = temp.read()
                if old_html != html_template:
                    with open(template_file_path, "w") as f:
                        f.write(html_template)
        return template_file


class SingleTags(Tags):
    left_delimiter = "<"
    right_delimiter = ">"
    is_single = True


class DoubleTags(Tags):
    left_delimiter = "<"
    right_delimiter = ">"


class SingleTemplates(SingleTags):
    left_delimiter = "{%"
    right_delimiter = "%}"
    self_dedent = False
    child_dedent = False
    enable_left_delimiter_space = True
    enable_right_delimiter_space = True
    enable_space_in_between = True

    def __init__(self,
                 template_name,
                 template_text,
                 *dom_elements,
                 self_dedent=None,
                 child_dedent=None,
                 **kwargs
                 ):
        open_tag = (
            "".join(
                [
                    self.left_delimiter,
                    " " if self.enable_left_delimiter_space else "",
                    template_name,
                    " " if self.enable_space_in_between else "",
                    template_text,
                    " " if self.enable_right_delimiter_space else "",
                    self.right_delimiter,
                ]
            )
            if any(template_name)
            else (
                "".join(
                    [
                        self.left_delimiter,
                        " " if self.enable_left_delimiter_space else "",
                        template_text,
                        " " if self.enable_right_delimiter_space else "",
                        self.right_delimiter
                    ]
                )
            )
        )
        super(SingleTemplates, self).__init__(open_tag=open_tag,
                                              self_dedent=self_dedent or self.self_dedent,
                                              child_dedent=child_dedent or self.child_dedent,
                                              **kwargs)
        if any(dom_elements):
            self.add(*dom_elements)


class DoubleTemplates(DoubleTags):
    left_delimiter = "{%"
    right_delimiter = "%}"
    self_dedent = True
    child_dedent = True
    closing_template_tag = "end"
    enable_left_delimiter_space = True
    enable_right_delimiter_space = True
    enable_space_in_between = True

    def __init__(self,
                 template_name,
                 template_text,
                 *dom_elements,
                 self_dedent=None,
                 child_dedent=None,
                 **kwargs
                 ):
        open_tag = "".join(
            [
                self.left_delimiter,
                " " if self.enable_left_delimiter_space else "",
                template_name,
                " " if self.enable_space_in_between else "",
                template_text,
                " " if self.enable_right_delimiter_space else "",
                self.right_delimiter,
            ]
        )
        close_tag = "".join([self.closing_template_tag, template_name])
        close_tag = "".join(
            [
                self.left_delimiter,
                " " if self.enable_left_delimiter_space else "",
                close_tag,
                " " if self.enable_right_delimiter_space else "",
                self.right_delimiter,
            ]
        )
        super(DoubleTemplates, self).__init__(open_tag=open_tag,
                                              close_tag=close_tag,
                                              self_dedent=self_dedent or self.self_dedent,
                                              child_dedent=child_dedent or self.child_dedent,
                                              **kwargs)
        if any(dom_elements):
            self.add(*dom_elements)


class StyleTags(Tags):
    left_delimiter = "{"
    right_delimiter = "}"
    self_dedent = False
    is_class = False
    is_id = False
    is_apply = False

    def _render_open_tag(self, sb, indent_level, indent_str, pretty, xhtml, name, open_tag):
        sb.append(name)
        if pretty:
            sb.append(" ")
        sb.append(self.left_delimiter)
        if pretty:
            sb.append("\n")
            sb.append(indent_str * indent_level)
        sb = self._render_attribute(sb, indent_level, indent_str, pretty, xhtml)
        # sb.append(''.join(["/", self.right_delimiter]) if self.is_single and xhtml else self.right_delimiter)
        return sb

    def _render_attribute(self, sb, indent_level, indent_str, pretty, xhtml):
        attribute_joiner = '%s:%s;' if not pretty else '    %s: %s;\n' + (
                indent_str * indent_level) if not self.is_apply else '@%s %s;' if not pretty else '    @%s %s;\n' + (
                indent_str * indent_level)
        for attribute, value in sorted(self.attributes.items()):
            if value is not False and value is not None:  # False values must be omitted completely
                sb.append(attribute_joiner % (attribute, escape(unicode(value), True)))

            if value is None:  # minified xhtml attributes are added
                sb.append(' %s' % attribute)
        return sb

    def _render_close_tag(self, sb, name, close_tag):
        sb.append(self.right_delimiter)
        return sb

    def _clean_name(self, name):
        # Workaround for python keywords and standard classes/methods
        # (del, object, input)
        if any(name):  # to handle the case when tagname = ""
            if name[-1] == '_':
                name = name[:-1]

        if all([self.is_class, self.is_id]):
            raise AttributeError(f"{name} can not have both is_class and is_id as True")

        if self.is_class:
            name = ''.join([".", name])

        if self.is_id:
            name = ''.join(["#", name])

        return name

    @property
    def attr(self):
        r = []
        attribute_joiner = '%s:%s;' if not self.is_apply else "@%s %s;"

        for attribute, value in sorted(self.attributes.items()):
            if value is not False and value is not None:  # False values must be omitted completely
                r.append(attribute_joiner % (attribute, escape(unicode(value), True)))

            if value is None:  # minified xhtml attributes are added
                r.append(' %s' % attribute)
        return u''.join(r)

    @staticmethod
    def clean_attribute(attribute):
        """
        Normalize attribute names for shorthand and work around for limitations
        in Python's syntax.
        """

        # Shorthand
        attribute = {
            'cls': 'class',
            'className': 'class',
            'class_name': 'class',
            'fr': 'for',
            'html_for': 'for',
            'htmlFor': 'for',
        }.get(attribute, attribute)

        # Workaround for Python's reserved words
        if attribute[0] == '_':
            attribute = attribute[1:]

        attribute = attribute.replace('_', '-')

        # Workaround for colon
        if attribute.split('_')[0] in ('xlink', 'xml', 'xmlns'):
            attribute = attribute.replace('_', ':', 1)

        return attribute


# if __name__ == '__main__':

    # class TestTag(Tags):
    #     pass
    #
    # class TestSingle(TestTag, SingleTags):
    #     pass
    #
    # class TestDouble(TestTag, DoubleTags):
    #     pass
    #
    # class SelfDedentTestSingle(TestSingle):
    #     self_dedent = True
    #
    # class SelfDedentTestDouble(TestDouble):
    #     self_dedent = True
    #
    # class DoubleSelfDedent(SelfDedentTestDouble):
    #     # is_inline = True
    #     pass
    #
    # class SingleSelfDedent(SelfDedentTestSingle):
    #     pass
    #
    #
    # tags = [
    #     Tags(Tags()),
    #     Tags(DoubleSelfDedent(Tags())),
    #     Tags(SingleSelfDedent(Tags())),
    #     Tags(SingleSelfDedent(SingleSelfDedent(Tags()))),
    #     Tags(DoubleSelfDedent(DoubleSelfDedent(Tags()))),
    #     Tags(SingleSelfDedent(DoubleSelfDedent(Tags()))),
    #     Tags(DoubleSelfDedent(SelfDedentTestDouble(Tags()))),
    #     Tags(SingleSelfDedent(Tags(SingleSelfDedent()))),
    #     Tags(DoubleSelfDedent(Tags(DoubleSelfDedent()))),
    #     Tags(SingleSelfDedent(Tags(DoubleSelfDedent()))),
    #     Tags(DoubleSelfDedent(Tags(SingleSelfDedent()))),
    #  ]
    #
    # for i, tag in enumerate(tags, 1):
    #     print(tag)

    # class TestTag(Tags):
    #     pass


    # class TestSingle(TestTag, SingleTags):
    #     pass


    # class TestDouble(TestTag, DoubleTags):
    #     pass


    # class SelfDedentTestSingle(TestSingle):
    #     self_dedent = True


    # class ChildDedentTestSingle(TestSingle):
    #     child_dedent = True


    # class SelfDedentTestDouble(TestDouble):
    #     self_dedent = True


    # class ChildDedentTestDouble(TestDouble):
    #     child_dedent = True


    # class DoubleSelfChildDedent(SelfDedentTestDouble, ChildDedentTestDouble):
    #     pass


    # class SingleSelfChildDedent(SelfDedentTestSingle, ChildDedentTestSingle):
    #     pass


    # class DoubleSelfDedent(SelfDedentTestDouble):
    #     # is_inline = True
    #     pass


    # class SingleSelfDedent(SelfDedentTestSingle):
    #     pass


    # class DoubleChildDedent(ChildDedentTestDouble):
    #     # render_tag = False
    #     pass


    # class SingleChildDedent(ChildDedentTestSingle):
    #     # render_tag = False
    #     pass


    # tags = [Tags(SingleSelfChildDedent(SingleSelfChildDedent(Tags()))),
    #         Tags(DoubleSelfChildDedent(DoubleSelfChildDedent(Tags()))),
    #         Tags(SingleSelfChildDedent(SingleSelfDedent(Tags()))),
    #         Tags(DoubleSelfChildDedent(DoubleSelfDedent("hb", Tags()))),
    #         Tags(SingleSelfChildDedent(SingleChildDedent(Tags()))),
    #         Tags(DoubleSelfChildDedent(DoubleChildDedent(Tags()))),
    #         Tags(SingleSelfChildDedent(Tags(SingleSelfChildDedent()))),
    #         Tags(DoubleSelfChildDedent(Tags(DoubleSelfChildDedent()))),
    #         Tags(SingleSelfChildDedent(Tags(SingleSelfDedent()))),
    #         Tags(DoubleSelfChildDedent(Tags(DoubleSelfDedent()))),
    #         Tags(SingleSelfChildDedent(Tags(SingleChildDedent()))),
    #         Tags(DoubleSelfChildDedent(Tags(DoubleChildDedent()))),
    #         Tags(SingleSelfDedent(SingleSelfChildDedent(Tags()))),
    #         Tags(DoubleSelfDedent(DoubleSelfChildDedent(Tags()))),
    #         Tags(SingleChildDedent(SingleSelfChildDedent(Tags()))),
    #         Tags(DoubleChildDedent(DoubleSelfChildDedent(Tags()))),
    #         Tags(SingleSelfDedent(SingleSelfDedent(Tags()))),
    #         Tags(DoubleSelfDedent(DoubleSelfDedent(Tags()))),
    #         Tags(SingleChildDedent(SingleChildDedent(Tags()))),
    #         Tags(DoubleChildDedent(DoubleChildDedent(Tags()))),
    #         ]

    # for i, tag in enumerate(tags, 1):
    #     print(i, "++++++++++++++++++++++++++++++++++++++++++++++++++++++")
    #     print(tag)
    #     print(i, "++++++++++++++++++++++++++++++++++++++++++++++++++++++")

    # x = Tags(Tags(x_bind_name="name"), x_for="name in names")
    # x["class"] = "a"
    # print(x)



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.htmlelement import HTMLElement


class OverflowAuto(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " overflow-auto"
        return elem


class OverflowHidden(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " overflow-hidden"
        return elem


class OverflowVisible(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " overflow-visible"
        return elem


class OverflowScroll(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " overflow-scroll"
        return elem


class OverflowXAuto(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " overflow-x-auto"
        return elem


class OverflowYAuto(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " overflow-y-auto"
        return elem


class OverflowXHidden(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " overflow-x-hidden"
        return elem


class OverflowYHidden(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " overflow-y-hidden"
        return elem


class OverflowXVisible(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " overflow-x-visible"
        return elem


class OverflowYVisible(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " overflow-y-visible"
        return elem


class OverflowXScroll(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " overflow-x-scroll"
        return elem


class OverflowYScroll(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " overflow-x-scroll"
        return elem



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

# Copyright (c) 2022 Valio
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT




# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.htmlelement import HTMLElement


class FlexShrink(HTMLElement):

    def __render__(self,  elem):
        elem["class"] += "flex-shrink"
        return elem


class FlexShrink0(HTMLElement):

    def __render__(self,  elem):
        elem["class"] += "flex-shrink-0"
        return elem



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from dataclasses import dataclass, field, make_dataclass
from typing import Union

from uidom.dom.src import DoubleTags, component

__all__ = [
    "HTMLElement",
    "AMPElement",
    "XComponent",
    "CustomElement",
    "WebComponent",
    "AlpineElement",
    "AlpineComponent",
]


@dataclass
class HtmlBaseMiddleware(object):
    """
    HTMLBaseMiddleware to intercept and modify dom elements.
    """


@dataclass(eq=False)
class HTMLElement(component.Component):
    file_extension = ".html"

    def __init__(self, *args, **kwargs):
        super(HTMLElement, self).__init__(*args, **kwargs)

    def __render__(self, *args, **kwargs):
        return self.html_tags.ConcatTag(*args, **kwargs)

    def __script__(self, *args, **kwargs):
        ...

    def __repr__(self):
        return str(self)


@dataclass(eq=False)
class AMPElement(HTMLElement):
    tag_name: str

    def __post_init__(self, *args, **kwargs):
        super(AMPElement, self).__init__(*args, tag_name=self.tag_name, **kwargs)

        class Element(DoubleTags):
            tagname = f"amp-{self.tag_name}"  # noqa

        element = Element

        self.Element = element
        self.Element.is_inline = self.is_inline
        self.Element.is_single = self.is_single
        self.Element.is_pretty = self.is_pretty

    def __call__(self, *args, **kwargs):
        return self.Element(*args, **kwargs)


# here init=False is necessary to avoid double initialization
@dataclass(init=False, eq=False)
class XComponent(HTMLElement):
    fields = None
    tag_name: str

    def __post_init__(self, *args, **kwargs):
        super(XComponent, self).__init__(*args, tag_name=self.tag_name, **kwargs)
        if self.fields is not None:
            element = make_dataclass(cls_name=''.join(map(lambda x: x.capitalize(), self.tag_name.split('-'))),
                                     fields=[*self.fields,
                                             ("tagname", str, field(init=False, default=f"x-{self.tag_name}")),
                                             ("attributes", dict, field(init=False, default_factory=lambda: dict())),
                                             ("children", list, field(init=False, default_factory=lambda: list())),
                                             ("document", list, field(init=False, default_factory=lambda: None)),
                                             ],
                                     bases=(DoubleTags,), )
        else:
            class Element(DoubleTags):
                tagname = f"x-{self.tag_name}"  # noqa

            element = Element

        self.Element = element
        self.Element.is_inline = self.is_inline
        self.Element.is_single = self.is_single
        self.Element.is_pretty = self.is_pretty
        # self.__call__ = element.__init__

    def __checks__(self, element):
        HTMLElement.__checks__(self, element)
        return self.__x_component_checks(element)

    def __x_component_checks(self, element):
        if not element.attributes.get("x-component"):
            if not element.get(x_component=self.tag_name):
                raise AttributeError(f"{element.__class__.__qualname__}: must have 'x_component' attribute")
        return super(HTMLElement, self).__checks__(element)

    def __call__(self, *args, **kwargs):
        return self.Element(*args, **kwargs)


@dataclass(eq=False)
class CustomElement(XComponent):
    # using x-component attribute to upgrade to the element in html to light-element

    def __checks__(self, element):
        XComponent.__checks__(self, element)
        return self.__custom_element_checks(element)

    def __custom_element_checks(self, element):  # noqa
        if (element.attributes.get("shadowroot") or element.get(shadowroot="shadowroot") or element.get(
                shadowroot="true")):
            raise AttributeError(f"{element.__class__.__qualname__}: must not have 'shadowroot' attribute")
        return element

    def __render__(self, tag_name):
        return self.html_tags.template(self.html_tags.div(x_component=tag_name))


@dataclass(eq=False)
class WebComponent(XComponent):
    # using x-component attribute to upgrade to the element in html to shadow-element

    def __checks__(self, element):
        XComponent.__checks__(self, element)
        return self.__web_component_checks(element)

    def __web_component_checks(self, element):  # noqa
        if not (element.attributes.get("shadowroot") or element.get(shadowroot="shadowroot") or element.get(
                shadowroot="true")):
            raise AttributeError(f"{element.__class__.__qualname__}: must have 'shadowroot' attribute set")
        return element

    def __render__(self, tag_name):
        # smallest example of a web component
        return self.html_tags.template(self.html_tags.div(x_component=tag_name, shadowroot="true"))


@dataclass(eq=False)
class AlpineElement(HTMLElement):
    x_data: Union[str, dict, bool, None] = False

    def __checks__(self, element):
        HTMLElement.__checks__(self, element)
        return self.__alpine_js_checks(element)

    def __alpine_js_checks(self, element):
        if not self.x_data:
            x_data = element.attributes.get("x-data") or element.get(x_data=self.x_data)
            if not x_data or x_data == "False":
                raise AttributeError(f"{element.__class__.__qualname__}: must have 'x-data' attribute")
        return element

    def __render__(self, *args, **kwargs):
        ...


@dataclass(eq=False)
class AlpineComponent(AlpineElement, XComponent):

    def __checks__(self, element):
        XComponent.__checks__(self, element)
        return AlpineElement.__checks__(self, element)


# if __name__ == '__main__':
    # image = CustomElement(tag_name="img")
    # print(image)
    # x = image(image(image("hello")), src="http://....", alt="some image", className="h-32 w-32")
    # print(x)
    # # print(signature(image))
    # def story(*args, **kwargs):
    #     return AMPElement(tag_name="story-page")(
    #         AMPElement(tag_name="story-grid-layer")(
    #             AMPElement(tag_name="video")(*args, **kwargs),
    #     template="fill"),
    #     id="cover"
    # )

    # print(
    #     story(
    #         layout="fill", 
    #         src="background.mp4", 
    #         poster="https://images.unsplash.com/photo-1605100804763-247f67b3557e?ixlib=rb-1.2.1&ixid=MnwxMjA3fDB8MHxwaG90by1wYWdlfHx8fGVufDB8fHx8&auto=format&fit=crop&w=870&q=80", 
    #         muted=None, 
    #         autoplay=None
    #         )
    #       )



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


from uidom.dom.htmlelement import HTMLElement


class Static(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " static"
        return elem


class Fixed(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " fixed"
        return elem


class Absolute(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " absolute"
        return elem


class Relative(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " relative"
        return elem


class Sticky(HTMLElement):

    def __render__(self, elem):
        elem["class"] += " sticky"
        return elem



# Copyright (c) 2022 uidom
# 
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


import builtins
import os
import sys
import types
import typing as T
from copy import deepcopy
from dataclasses import dataclass, field

from uidom.dom.htmlelement import HTMLElement
from uidom.dom.src import ext, htmltags, jinjatags, svgtags
from uidom.dom.src.htmltags import ConcatTag, html_tag
from uidom.dom.src.parse_html import Element, tokenize_html
from uidom.dom.src.utils import dom_text

__all__ = [
    "HTMLStringToDom",
    "HTMLStringToElement"
]


def create_dynamic_element(tag_name: str) -> T.Type[ext.Tags]:
    class _Element(ext.Tags):
        tagname = tag_name

    cls_name = ''.join(map(lambda x: x.capitalize(), tag_name.split("-")))
    _Element.__qualname__ = cls_name
    _Element.__name__ = cls_name
    module_name: str = os.path.splitext(os.path.basename(sys.modules["__main__"].__file__))[0] # type: ignore
    _Element.__module__ = module_name
    element: T.Type[ext.Tags]  = deepcopy(_Element)
    del _Element
    setattr(sys.modules["__main__"], cls_name, element)
    return element


@dataclass
class HTMLStringToDom(object):
    html_string_or_token: T.Union[str, Element, list[Element], ext.Tags]
    modules: list[types.ModuleType] = field(default_factory=list)

    def __post_init__(self):
        self.tokens: T.Union[Element, list[Element], ext.Tags] = tokenize_html(
            self.html_string_or_token if not isinstance(self.html_string_or_token, html_tag)
            else str(self.html_string_or_token)
        ).children if isinstance(self.html_string_or_token, (str, html_tag)) else self.html_string_or_token
        
        # for htmlt, svg and jinja tags lookup 
        if htmltags not in self.modules:
            self.modules.append(htmltags)
            
        if svgtags not in self.modules:
            self.modules.append(svgtags)
            
        if jinjatags not in self.modules:
            self.modules.append(jinjatags)
        
    def parse(self, tag: T.Optional[ext.Tags] = None) -> T.Union[str, ext.Tags, None]:
        for token in self.tokens:
            if not token.name:
                # myst_parser gives out Data['abc'] kind of Token with {"name":'', 'data': 'abc', ...} attributes
                # we are parsing plain strings, <script> body or jinja token strings here

                if hasattr(token, "data"):
                    # handle leading and trailing newline with spaces in myst parser Data Token ex: "  \n Hello\n  "
                    if data := token.data.strip("\n").strip(" ").strip("\n"):
                        with (tag or ConcatTag()) as tag:
                            # handling
                            if tag.__class__.__name__ == "script":
                                dom_text(data, escape=False)
                            else:
                                if token.__class__.__name__ != "Comment":
                                    if data.startswith("{") and data.endswith("}"):
                                        # these are jinja tags
                                        dom_text(data, escape=False)
                                    else:
                                        # normal text is here
                                        dom_text(data, escape=True)
                                else:
                                    # this is a comment section
                                    htmltags.comment(data)

            else:
                element = None
                tag_name = token.name
                if tag_name in builtins.__dict__:
                    # work around for builtins like 'input' tag
                    tag_name = ''.join([token.name, "_"])
                    
                for module in self.modules:
                    try:
                        element = getattr(module, tag_name)
                    except (AttributeError,):
                        pass
                if element is None:
                    element = create_dynamic_element(tag_name=tag_name)
                
                tag = tag if tag is not None else ConcatTag()
                
                with tag:
                    with element(**token.attrs) as child_tag:
                        HTMLStringToDom(token.children, modules=self.modules).parse(tag=child_tag)
        return tag

    def __repr__(self):
        return str(self.parse())


@dataclass
class HTMLStringToElement(HTMLElement):
    # TODO convert DOM object instance to Code Object for AST conversion to python code.
    # from https://stackoverflow.com/questions/68577587/how-to-find-the-ast-assignment-node-related-to-the-instance-creation
    # this.ast_object can be easily used to create python code for any html object
    # https://stackoverflow.com/a/68584740 for parsing a python object into an ast_object
    # https://stackoverflow.com/a/63212256 for unparsing a python ast_object


    def __init__(self, *args, **kwargs):
        super(HTMLStringToElement, self).__init__(*args, **kwargs)

    def __render__(self, raw_string) -> T.Union[str, ext.Tags, None]: # noqa
        return HTMLStringToDom(raw_string).parse()


# if __name__ == '__main__':
    # from uidom.dom import ConcatTag, For, Var, div, li, raw, script, ul

    # print(HTMLStringToElement("<li><ul><i><!--Hello World--></i></ul><a href='www.google.com'></a></li>"))
#     class XName(ext.Tags):
#         tagname = "x-name"


    # print(HTMLStringToDom(div("hello", div("Jai SHree Ram"), script(raw("function () => {}")), className="sdaf")).parse())
    # print(div("hello", div("Jai SHree Ram"), script(raw("function () => {}")), className="sdaf"))
    # print(StringToDom(str(div("hello", div("Jai SHree Ram"), script(raw("function () => {}")), className="sdaf", x_data=None))))
    # x = HTMLToPy(str(XName("hello", Var("haha"), ul(For("name in names", li(Var("name")))),
    #                        div("Jai SHree Ram aa", className="safn"), script(raw("function () => {}")),
    #                         script(src="https://unpkg.com/filepond/dist/filepond.js"),
    #                         className="sdaf", x_data={}, x_transition_enter="")))
    # print(x)
    # print(XName("hello >", Var("haha >"), ul(For("name in names", li(Var("name")))),
    #             div("Jai SHree Ram   a ", className="safn"), script(raw("function () => {}")),
    #                         script(src="https://unpkg.com/filepond/dist/filepond.js"),
    #                         className="sdaf", x_data={}, x_transition_enter=""))
    # print(HTMLStringToDom(str(ul(For("name in names", li(Var("name")))))))
    
